from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode, decode
from ..._internal import integration_manager as integration_manager_lib
from ...utils.encoding import encoding_to_types
from typing import Any, TypedDict

# --------------------------------------------------------------------------------------------
# TAKE ORDER
# --------------------------------------------------------------------------------------------


async def take_order(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, Any]
            {
                "expected_incoming_asset_amount": int,
                "min_incoming_asset_amount": int,
                "outgoing_asset": ChecksumAddress,
                "outgoing_asset_amount": int,
                "uuid": HexStr,
                "swap_type": int,
                "swap_data": Any,
            }
    """
    _take_order = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["take_order"], take_order_encode
    )
    return await _take_order(*args)


TAKE_ORDER_ENCODING = [
    {
        "name": "minIncomingAssetAmount",
        "type": "uint256",
    },
    {
        "name": "expectedIncomingAssetAmount",
        "type": "uint256",
    },
    {
        "name": "outgoingAsset",
        "type": "address",
    },
    {
        "name": "outgoingAssetAmount",
        "type": "uint256",
    },
    {
        "name": "uuid",
        "type": "bytes16",
    },
    {
        "name": "swapType",
        "type": "uint256",
    },
    {
        "name": "swapData",
        "type": "bytes",
    },
]

ROUTE_ENCODING = {
    "components": [
        {
            "name": "index",
            "type": "uint256",
        },
        {
            "name": "targetExchange",
            "type": "address",
        },
        {
            "name": "percent",
            "type": "uint256",
        },
        {
            "name": "payload",
            "type": "bytes",
        },
        {
            "name": "networkFee",
            "type": "uint256",
        },
    ]
}

ADAPTER_ENCODING = {
    "components": [
        {
            "name": "adapter",
            "type": "address",
        },
        {
            "name": "percent",
            "type": "uint256",
        },
        {
            "name": "networkFee",
            "type": "uint256",
        },
        {
            "name": "route",
            "type": "tuple[]",
            "components": ROUTE_ENCODING["components"],
        },
    ],
}

PATH_ENCODING = {
    "components": [
        {
            "name": "to",
            "type": "address",
        },
        {
            "name": "totalNetworkFee",
            "type": "uint256",
        },
        {
            "name": "adapters",
            "type": "tuple[]",
            "components": ADAPTER_ENCODING["components"],
        },
    ],
}

MEGA_SWAP_DATA_ENCODING = [
    {
        "components": [
            {
                "name": "fromAmountPercent",
                "type": "uint256",
            },
            {
                "name": "path",
                "type": "tuple[]",
                "components": PATH_ENCODING["components"],
            },
        ],
        "name": "megaSwapData",
        "type": "tuple[]",
    }
]

MULTI_SWAP_DATA_ENCODING = [
    {
        "name": "multiSwapPath",
        "type": "tuple[]",
        "components": PATH_ENCODING["components"],
    }
]

SIMPLE_SWAP_DATA_ENCODING = [
    {
        "components": [
            {
                "name": "incomingAsset",
                "type": "address",
            },
            {
                "name": "callees",
                "type": "address[]",
            },
            {
                "name": "exchangeData",
                "type": "bytes",
            },
            {
                "name": "startIndexes",
                "type": "uint256[]",
            },
            {
                "name": "values",
                "type": "uint256[]",
            },
        ],
        "name": "simpleSwapParams",
        "type": "tuple",
    }
]


class Route(TypedDict):
    index: int
    targetExchange: ChecksumAddress
    percent: int
    payload: HexStr
    networkFee: int


class Adapter(TypedDict):
    adapter: ChecksumAddress
    percent: int
    networkFee: int
    route: list[Route]


class Path(TypedDict):
    to: ChecksumAddress
    totalNetworkFee: int
    adapters: list[Adapter]


class MegaSwapPathData(TypedDict):
    fromAmountPercent: int
    path: list[Path]


MegaSwapData = list[MegaSwapPathData]
MultiSwapData = list[Path]


class SimpleSwapData(TypedDict):
    incomingAsset: ChecksumAddress
    callees: list[ChecksumAddress]
    exchangeData: HexStr
    startIndexes: list[int]
    values: list[int]


class SwapType(TypedDict):
    simple: int
    multi: int
    mega: int


SWAP_TYPE = {
    "simple": 0,
    "multi": 1,
    "mega": 2,
}


class TakeOrderArgs(TypedDict):
    expected_incoming_asset_amount: int
    min_incoming_asset_amount: int
    outgoing_asset: ChecksumAddress
    outgoing_asset_amount: int
    uuid: HexStr
    swap_type: int
    swap_data: MegaSwapData | MultiSwapData | SimpleSwapData


def take_order_encode(
    expected_incoming_asset_amount: int,
    min_incoming_asset_amount: int,
    outgoing_asset: ChecksumAddress,
    outgoing_asset_amount: int,
    uuid: HexStr,
    swap_type: int,
    swap_data: Any,
) -> HexStr:
    """
    Args:
        if swap_type == 0 (simple):
            swap_data:
                {
                    "incomingAsset": ChecksumAddress,
                    "callees": list[ChecksumAddress],
                    "exchangeData": HexStr,
                    "startIndexes": list[int],
                    "values": list[int],
                }
        elif swap_type == 1 (multi):
            swap_data:
                [
                    {
                        "to": ChecksumAddress,
                        "totalNetworkFee": int,
                        "adapters":
                            [
                                {
                                    "adapter": ChecksumAddress,
                                    "percent": int,
                                    "networkFee": int,
                                    "route":
                                        [
                                            {
                                                "index": int,
                                                "targetExchange": ChecksumAddress,
                                                "percent": int,
                                                "payload": HexStr,
                                                "networkFee": int,
                                            }
                                        ]
                                }
                            ]
                    }
                ]
        elif swap_type == 2 (mega):
            swap_data:
                [
                    {
                        "fromAmountPercent": int,
                        "path":
                            [
                                {
                                    "to": ChecksumAddress,
                                    "totalNetworkFee": int,
                                    "adapters":
                                        [
                                            {
                                                "adapter": ChecksumAddress,
                                                "percent": int,
                                                "networkFee": int,
                                                "route":
                                                    [
                                                        {
                                                            "index": int,
                                                            "targetExchange": ChecksumAddress,
                                                            "percent": int,
                                                            "payload": HexStr,
                                                            "networkFee": int,
                                                        }
                                                    ]
                                            }
                                        ]
                                }
                            ]
                    }
                ]
    """

    def _extract_paths_values(paths):
        paths_values = []
        for path in paths:
            adapters_values = []
            for adapter in path["adapters"]:
                route_values = []
                for route in adapter["route"]:
                    route_values.append(
                        (
                            int(route["index"]),
                            Web3.to_checksum_address(route["targetExchange"]),
                            int(route["percent"]),
                            Web3.to_bytes(hexstr=route["payload"]),
                            int(route["networkFee"]),
                        )
                    )
                adapters_values.append(
                    (
                        Web3.to_checksum_address(adapter["adapter"]),
                        int(adapter["percent"]),
                        int(adapter["networkFee"]),
                        route_values,
                    )
                )
            paths_values.append(
                (
                    Web3.to_checksum_address(path["to"]),
                    int(path["totalNetworkFee"]),
                    adapters_values,
                )
            )
        return paths_values

    if swap_type == SWAP_TYPE["mega"]:
        swap_data_types = encoding_to_types(MEGA_SWAP_DATA_ENCODING)
        swap_data_values = [
            [
                (int(part["fromAmountPercent"]), _extract_paths_values(part["path"]))
                for part in swap_data
            ]
        ]
    elif swap_type == SWAP_TYPE["multi"]:
        swap_data_types = encoding_to_types(MULTI_SWAP_DATA_ENCODING)
        swap_data_values = [_extract_paths_values(swap_data)]
    elif swap_type == SWAP_TYPE["simple"]:
        swap_data_types = encoding_to_types(SIMPLE_SWAP_DATA_ENCODING)
        swap_data_values = [
            (
                swap_data["incoming_asset"],
                swap_data["callees"],
                Web3.to_bytes(hexstr=swap_data["exchange_data"]),
                swap_data["start_indexes"],
                swap_data["values"],
            ),
        ]
    else:
        raise ValueError(f"Invalid swap_type: {swap_type}")

    encoded_swap_data = encode(swap_data_types, swap_data_values)

    types = encoding_to_types(TAKE_ORDER_ENCODING)
    values = [
        min_incoming_asset_amount,
        expected_incoming_asset_amount,
        outgoing_asset,
        outgoing_asset_amount,
        Web3.to_bytes(hexstr=uuid),
        swap_type,
        encoded_swap_data,
    ]
    return Web3.to_hex(encode(types, values))


def take_order_decode(encoded: HexStr) -> dict[str, Any]:
    """
    Returns:
        dict[str, Any]
            {
                "min_incoming_asset_amount": int,
                "expected_incoming_asset_amount": int,
                "outgoing_asset": ChecksumAddress,
                "outgoing_asset_amount": int,
                "uuid": HexStr,
                "swap_type": int,
                "swap_data": Any,
            }
    """
    types = encoding_to_types(TAKE_ORDER_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))

    swap_type = decoded[5]
    swap_data_encoded = decoded[6]

    def _rebuild_paths_values(paths_values):
        paths = []
        for path in paths_values:
            adapters = []
            for adapter in path[2]:
                routes = []
                for route in adapter[3]:
                    routes.append(
                        {
                            "index": route[0],
                            "targetExchange": Web3.to_checksum_address(route[1]),
                            "percent": route[2],
                            "payload": Web3.to_hex(route[3]),
                            "networkFee": route[4],
                        }
                    )
                adapters.append(
                    {
                        "adapter": Web3.to_checksum_address(adapter[0]),
                        "percent": adapter[1],
                        "networkFee": adapter[2],
                        "route": routes,
                    }
                )
            paths.append(
                {
                    "to": Web3.to_checksum_address(path[0]),
                    "totalNetworkFee": path[1],
                    "adapters": adapters,
                }
            )
        return paths

    if swap_type == SWAP_TYPE["mega"]:
        swap_data_types = encoding_to_types(MEGA_SWAP_DATA_ENCODING)
        (swap_data_decoded,) = decode(swap_data_types, swap_data_encoded)
        swap_data = [
            {
                "fromAmountPercent": part[0],
                "path": _rebuild_paths_values(part[1]),
            }
            for part in swap_data_decoded
        ]
    elif swap_type == SWAP_TYPE["multi"]:
        swap_data_types = encoding_to_types(MULTI_SWAP_DATA_ENCODING)
        (swap_data_decoded,) = decode(swap_data_types, swap_data_encoded)
        swap_data = _rebuild_paths_values(swap_data_decoded)
    elif swap_type == SWAP_TYPE["simple"]:
        swap_data_types = encoding_to_types(SIMPLE_SWAP_DATA_ENCODING)
        (swap_data_decoded,) = decode(swap_data_types, swap_data_encoded)
        swap_data = {
            "incoming_asset": swap_data_decoded[0],
            "callees": list(swap_data_decoded[1]),
            "exchange_data": Web3.to_hex(swap_data_decoded[2]),
            "start_indexes": list(swap_data_decoded[3]),
            "values": list(swap_data_decoded[4]),
        }
    else:
        raise ValueError(f"Invalid swap_type: {swap_type}")

    return {
        "min_incoming_asset_amount": decoded[0],
        "expected_incoming_asset_amount": decoded[1],
        "outgoing_asset": decoded[2],
        "outgoing_asset_amount": decoded[3],
        "uuid": Web3.to_hex(decoded[4]),
        "swap_type": swap_type,
        "swap_data": swap_data,
    }


# --------------------------------------------------------------------------------------------
# TAKE MULTIPLE ORDERS
# --------------------------------------------------------------------------------------------


async def take_multiple_orders(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, Any]
            {
                "allow_orders_to_fail": bool,
                "orders": list[dict[str, Any]],
            }
    """
    _take_multiple_orders = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["take_multiple_orders"],
        take_multiple_orders_encode,
    )
    return await _take_multiple_orders(*args)


TAKE_MULTIPLE_ORDERS_ENCODING = [
    {
        "name": "ordersData",
        "type": "bytes[]",
    },
    {
        "name": "allowOrdersToFail",
        "type": "bool",
    },
]


class TakeMultipleOrdersArgs(TypedDict):
    allow_orders_to_fail: bool
    orders: list[TakeOrderArgs]


def take_multiple_orders_encode(
    allow_orders_to_fail: bool,
    orders: list[TakeOrderArgs],
) -> HexStr:
    """
    Args:
        orders: list[dict[str, Any]]
            [
                {
                    "expected_incoming_asset_amount": int,
                    "min_incoming_asset_amount": int,
                    "outgoing_asset": ChecksumAddress,
                    "outgoing_asset_amount": int,
                    "uuid": HexStr,
                    "swap_type": int,
                    "swap_data": Any,
                }
            ]
    """
    orders_data = [Web3.to_bytes(hexstr=take_order_encode(**order)) for order in orders]
    types = encoding_to_types(TAKE_MULTIPLE_ORDERS_ENCODING)
    values = [orders_data, allow_orders_to_fail]
    return Web3.to_hex(encode(types, values))


def take_multiple_orders_decode(encoded: HexStr) -> dict[str, Any]:
    """
    Returns:
        dict[str, Any]
            {
                "orders": list[dict[str, Any]],
                "allow_orders_to_fail": bool,
            }
    """
    types = encoding_to_types(TAKE_MULTIPLE_ORDERS_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))

    return {
        "orders": [take_order_decode(Web3.to_hex(order)) for order in decoded[0]],
        "allow_orders_to_fail": decoded[1],
    }



from ..releases import RELEASES
from ..assets import define_asset_list
from ..networks import NETWORK_BY_SLUG
from ..price_feeds import RateAsset


SULU = RELEASES["base"]["sulu"]


ASSETS = define_asset_list(
    NETWORK_BY_SLUG["base"],
    [
        {
            "decimals": 18,
            "id": "0x4200000000000000000000000000000000000006",
            "name": "Wrapped Ether",
            "releases": [SULU],
            "symbol": "WETH",
            "type": "primitive",
            "price_feed": {
                "type": "WETH",
            },
        },
        {
            "decimals": 18,
            "id": "0x7C298664BD6582f6f264c2Cb5a4B9cC09b6E3889",
            "name": "Melon Token",
            "releases": [SULU],
            "symbol": "MLN",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x122b5334A8b55861dBc6729c294451471FbF318D",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x50c5725949A6F0c72E6C4a641F24049A917DB0Cb",
            "name": "Dai Stablecoin",
            "releases": [SULU],
            "symbol": "DAI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x591e79239a7d679378eC8c847e5038150364C78F",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x9e1028F5F1D5eDE59748FFceE5532509976840E0",
            "name": "Compound",
            "releases": [SULU],
            "symbol": "COMP",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x9DDa783DE64A9d1A60c49ca761EbE528C35BA428",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
            "name": "USD Coin",
            "releases": [SULU],
            "symbol": "USDC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x7e860098F58bBFC8648a4311b374B1D669a2bc6B",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xfde4C96c8593536E31F229EA8f37b2ADa2699bb2",
            "name": "Tether USD",
            "releases": [SULU],
            "symbol": "USDT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xf19d560eB8d2ADf07BD6D13ed03e1D11215721F9",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 8,
            "id": "0x0555E30da8f98308EdB960aa94C0Db47230d2B9c",
            "name": "Wrapped BTC",
            "releases": [SULU],
            "symbol": "WBTC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xCCADC697c55bbB68dc5bCdf8d3CBe83CdD4E071E",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xD4a0e0b9149BCee3C920d2E00b5dE09138fd8bb7",
            "name": "Aave Base WETH",
            "releases": [SULU],
            "symbol": "aBasWETH",
            "type": "aave-v3",
            "underlying": "0x4200000000000000000000000000000000000006",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x71041dddad3595F9CEd3DcCFBe3D1F4b0a16Bb70",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x4e65fE4DbA92790696d040ac24Aa414708F5c0AB",
            "name": "Aave Base USDC",
            "releases": [SULU],
            "symbol": "aBasUSDC",
            "type": "aave-v3",
            "underlying": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x7e860098F58bBFC8648a4311b374B1D669a2bc6B",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "id": "0xbeeF010f9cb27031ad51e3333f9aF9C6B1228183",
            "decimals": 18,
            "name": "Steakhouse USDC",
            "symbol": "steakUSDC",
            "releases": [SULU],
            "type": "erc-4626",
            "underlying": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
            "protocol": "morpho",
            "price_feed": {
                "type": "DERIVATIVE_ERC4626",
                "address": "0x6889790Fb10A03bBf9dc86f1BEd3219b509F5367",
            },
        },
        {
            "id": "0xbEEf050a7485865A7a8d8Ca0CC5f7536b7a3443e",
            "decimals": 18,
            "name": "Steakhouse ETH",
            "symbol": "steakETH",
            "releases": [SULU],
            "type": "erc-4626",
            "underlying": "0x4200000000000000000000000000000000000006",
            "protocol": "morpho",
            "price_feed": {
                "type": "DERIVATIVE_ERC4626",
                "address": "0x6889790Fb10A03bBf9dc86f1BEd3219b509F5367",
            },
        },
        {
            "decimals": 8,
            "id": "0xcbB7C0000aB88B473b1f5aFd9ef808440eed33Bf",
            "name": "Coinbase Wrapped BTC",
            "releases": [SULU],
            "symbol": "cbBTC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x07DA0E54543a844a80ABE69c8A12F22B3aA59f9D",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xc1CBa3fCea344f92D9239c08C0568f6F2F0ee452",
            "name": "Wrapped liquid staked Ether 2.0",
            "releases": [SULU],
            "symbol": "wstETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x43a5C292A453A3bF3606fa856197f09D7B74251a",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x99CBC45ea5bb7eF3a5BC08FB1B7E56bB2442Ef0D",
            "name": "Aave Base wstETH",
            "releases": [SULU],
            "symbol": "aBaswstETH",
            "type": "aave-v3",
            "underlying": "0xc1CBa3fCea344f92D9239c08C0568f6F2F0ee452",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x43a5C292A453A3bF3606fa856197f09D7B74251a",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x1Bc71130A0e39942a7658878169764Bbd8A45993",
            "name": "KelpDao Restaked ETH",
            "releases": [SULU],
            "symbol": "rsETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xd7221b10FBBC1e1ba95Fd0B4D031C15f7F365296",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xDBFeFD2e8460a6Ee4955A68582F85708BAEA60A3",
            "name": "Super OETH",
            "symbol": "superOETHb",
            "releases": [SULU],
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x71041dddad3595F9CEd3DcCFBe3D1F4b0a16Bb70",
                "rate_asset": RateAsset.USD,
                "non_standard": True,
                "pegged_to": "ETH",
            },
        },
        {
            "id": "0x7FcD174E80f264448ebeE8c88a7C4476AAF58Ea6",
            "name": "Wrapped Super OETH",
            "symbol": "wsuperOETHb",
            "decimals": 18,
            "releases": [SULU],
            "type": "primitive",
            "price_feed": {
                "type": "DERIVATIVE_ERC4626",
                "address": "0x6889790Fb10A03bBf9dc86f1BEd3219b509F5367",
                "non_standard": True,
            },
        },
    ],
)



from typing import Callable, Any, TypedDict
from web3.types import ChecksumAddress, HexStr, TxParams
from web3.constants import ADDRESS_ZERO
from eth_abi import encode, decode
from web3 import Web3
from ..utils.clients import WalletClient
from ..utils.encoding import encoding_to_types
from .extensions import call_extension


class Action(TypedDict):
    create_external_position: int
    call_on_external_position: int
    remove_external_position: int
    reactivate_external_position: int


ACTION = {
    "create_external_position": 0,
    "call_on_external_position": 1,
    "remove_external_position": 2,
    "reactivate_external_position": 3,
}


class UseParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    external_position_proxy: ChecksumAddress
    call_args: dict[str, Any]


def make_use(action: int, encoder: Callable | None = None) -> Callable:
    async def use_external_position(
        client: WalletClient,
        comptroller_proxy: ChecksumAddress,
        external_position_manager: ChecksumAddress,
        external_position_proxy: ChecksumAddress,
        call_args: dict[str, Any],
    ) -> TxParams:
        return await call(
            client,
            comptroller_proxy,
            external_position_manager,
            external_position_proxy,
            action,
            "0x" if encoder is None else encoder(**call_args),
        )

    return use_external_position


class CreateAndUseParams(TypedDict):
    client: WalletClient
    type_id: int
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    initialization_data: HexStr
    call_args: dict[str, Any]


def make_create_and_use(action: int, encoder: Callable | None = None) -> Callable:
    async def create_and_use(
        client: WalletClient,
        type_id: int,
        comptroller_proxy: ChecksumAddress,
        external_position_manager: ChecksumAddress,
        initialization_data: HexStr,
        call_args: dict[str, Any],
    ) -> TxParams:
        return await create(
            client,
            type_id,
            comptroller_proxy,
            external_position_manager,
            initialization_data,
            call_encode(
                ADDRESS_ZERO,
                action,
                "0x" if encoder is None else encoder(**call_args),
            ),
        )

    return create_and_use


# --------------------------------------------------------------------------------------------
# CALL ON EXTERNAL POSITION
# --------------------------------------------------------------------------------------------

CALL_ENCODING = [
    {
        "name": "externalPositionProxy",
        "type": "address",
    },
    {
        "name": "actionId",
        "type": "uint256",
    },
    {
        "name": "actionArgs",
        "type": "bytes",
    },
]


class CallArgs(TypedDict):
    external_position_proxy: ChecksumAddress
    action_id: int
    action_args: HexStr | None


def call_encode(
    external_position_proxy: ChecksumAddress,
    action_id: int,
    action_args: HexStr | None = None,
) -> HexStr:
    types = encoding_to_types(CALL_ENCODING)
    values = [
        external_position_proxy,
        action_id,
        Web3.to_bytes(hexstr=action_args or "0x"),
    ]
    return Web3.to_hex(encode(types, values))


def call_decode(encoded: HexStr) -> dict[str, ChecksumAddress | int | HexStr]:
    """
    Returns:
        {
            "external_position_proxy": ChecksumAddress,
            "action_id": int,
            "action_args": HexStr,
        }
    """
    types = encoding_to_types(CALL_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "external_position_proxy": Web3.to_checksum_address(decoded[0]),
        "action_id": decoded[1],
        "action_args": Web3.to_hex(decoded[2]),
    }


class CallParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    external_position_proxy: ChecksumAddress
    action_id: int
    action_args: HexStr


async def call(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    external_position_manager: ChecksumAddress,
    external_position_proxy: ChecksumAddress,
    action_id: int,
    action_args: HexStr,
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        external_position_manager,
        ACTION["call_on_external_position"],
        call_encode(external_position_proxy, action_id, action_args),
    )


# --------------------------------------------------------------------------------------------
# CREATE EXTERNAL POSITION
# --------------------------------------------------------------------------------------------

CREATE_ENCODING = [
    {
        "name": "typeId",
        "type": "uint256",
    },
    {
        "name": "initializationData",
        "type": "bytes",
    },
    {
        "name": "callOnExternalPositionCallArgs",
        "type": "bytes",
    },
]


class CreateArgs(TypedDict):
    type_id: int
    initialization_data: HexStr | None
    call_on_external_position_call_args: HexStr | None


def create_encode(
    type_id: int,
    initialization_data: HexStr | None,
    call_on_external_position_call_args: HexStr | None,
) -> HexStr:
    types = encoding_to_types(CREATE_ENCODING)
    values = [
        type_id,
        Web3.to_bytes(hexstr=initialization_data or "0x"),
        Web3.to_bytes(hexstr=call_on_external_position_call_args or "0x"),
    ]
    return Web3.to_hex(encode(types, values))


def create_decode(encoded: HexStr) -> dict[str, int | HexStr | HexStr]:
    """
    Returns:
        {
            "type_id": int,
            "initialization_data": HexStr,
            "call_on_external_position_call_args": HexStr,
        }
    """
    types = encoding_to_types(CREATE_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "type_id": decoded[0],
        "initialization_data": Web3.to_hex(decoded[1]),
        "call_on_external_position_call_args": Web3.to_hex(decoded[2]),
    }


class CreateParams(TypedDict):
    client: WalletClient
    type_id: int
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    initialization_data: HexStr | None
    call_args: HexStr | None


async def create(
    client: WalletClient,
    type_id: int,
    comptroller_proxy: ChecksumAddress,
    external_position_manager: ChecksumAddress,
    initialization_data: HexStr | None,
    call_args: HexStr | None,
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        external_position_manager,
        ACTION["create_external_position"],
        create_encode(type_id, initialization_data or "0x", call_args or "0x"),
    )


async def create_only(
    client: WalletClient,
    type_id: int,
    comptroller_proxy: ChecksumAddress,
    external_position_manager: ChecksumAddress,
    initialization_data: HexStr | None,
) -> TxParams:
    return await create(*locals().values(), None)


# --------------------------------------------------------------------------------------------
# REACTIVATE EXTERNAL POSITION
# --------------------------------------------------------------------------------------------

REACTIVATE_ENCODING = [
    {
        "name": "externalPosition",
        "type": "address",
    },
]


class ReactivateArgs(TypedDict):
    external_position: ChecksumAddress


def reactivate_encode(external_position: ChecksumAddress) -> HexStr:
    types = encoding_to_types(REACTIVATE_ENCODING)
    values = [external_position]
    return Web3.to_hex(encode(types, values))


def reactivate_decode(encoded: HexStr) -> dict[str, ChecksumAddress]:
    """
    Returns:
        {
            "external_position": ChecksumAddress,
        }
    """
    types = encoding_to_types(REACTIVATE_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "external_position": Web3.to_checksum_address(decoded[0]),
    }


class ReactivateParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    external_position_proxy: ChecksumAddress


async def reactivate(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    external_position_manager: ChecksumAddress,
    external_position_proxy: ChecksumAddress,
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        external_position_manager,
        ACTION["reactivate_external_position"],
        reactivate_encode(external_position_proxy),
    )


# --------------------------------------------------------------------------------------------
# REMOVE EXTERNAL POSITION
# --------------------------------------------------------------------------------------------

REMOVE_ENCODING = [
    {
        "name": "externalPositionProxy",
        "type": "address",
    },
]


class RemoveArgs(TypedDict):
    external_position_proxy: ChecksumAddress


def remove_encode(external_position_proxy: ChecksumAddress) -> HexStr:
    types = encoding_to_types(REMOVE_ENCODING)
    values = [external_position_proxy]
    return Web3.to_hex(encode(types, values))


def remove_decode(encoded: HexStr) -> dict[str, ChecksumAddress]:
    """
    Returns:
        {
            "external_position_proxy": ChecksumAddress,
        }
    """
    types = encoding_to_types(REMOVE_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "external_position_proxy": Web3.to_checksum_address(decoded[0]),
    }


class RemoveParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    external_position_manager: ChecksumAddress
    external_position_proxy: ChecksumAddress


async def remove(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    external_position_manager: ChecksumAddress,
    external_position_proxy: ChecksumAddress,
) -> TxParams:
    return await call_extension(
        client,
        comptroller_proxy,
        external_position_manager,
        ACTION["remove_external_position"],
        remove_encode(external_position_proxy),
    )



from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode
from ...utils.clients import PublicClient, WalletClient
from ...utils.encoding import encoding_to_types
from ....abis import abis

# --------------------------------------------------------------------------------------------
# FACTORY
# --------------------------------------------------------------------------------------------


async def deploy_proxy(
    client: WalletClient,
    share_price_throttled_asset_manager_factory: ChecksumAddress,
    owner: ChecksumAddress,
    vault_proxy_address: ChecksumAddress,
    loss_tolerance: int,
    loss_tolerance_period_duration: int,
    shutdowner: ChecksumAddress,
) -> TxParams:
    contract = client.contract(
        share_price_throttled_asset_manager_factory,
        abis.ISharePriceThrottledAssetManagerFactory,
    )
    function = contract.functions.deployProxy(
        owner,
        vault_proxy_address,
        loss_tolerance,
        loss_tolerance_period_duration,
        shutdowner,
    )
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# LIB
# --------------------------------------------------------------------------------------------


EXECUTE_CALLS_ENCODING = [
    {
        "type": "tuple[]",
        "name": "calls",
        "components": [
            {
                "name": "target",
                "type": "address",
            },
            {
                "name": "data",
                "type": "bytes",
            },
        ],
    },
]


async def execute_calls(
    client: WalletClient,
    share_price_throttled_asset_manager: ChecksumAddress,
    calls: list[ChecksumAddress, HexStr],
) -> TxParams:
    """
    Args:
        client: WalletClient
        share_price_throttled_asset_manager: ChecksumAddress
        calls: list[ChecksumAddress, HexStr]
            [
                [
                    ChecksumAddress,  # target
                    HexStr,           # data
                ],
                ...
            ]
    """
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    types = encoding_to_types(EXECUTE_CALLS_ENCODING)
    args = [[[target, Web3.to_bytes(hexstr=data)] for target, data in calls]]
    encoded_calls = Web3.to_hex(encode(types, args))
    function = contract.functions.executeCalls(encoded_calls)
    return await client.populated_transaction(function)


async def shutdown(
    client: WalletClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> TxParams:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.shutdown()
    return await client.populated_transaction(function)


async def get_loss_tolerance(
    client: PublicClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> int:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.getLossTolerance()
    return await function.call()


async def get_loss_tolerance_period_duration(
    client: PublicClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> int:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.getLossTolerancePeriodDuration()
    return await function.call()


async def get_shutdowner(
    client: PublicClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.getShutdowner()
    return await function.call()


async def get_throttle(
    client: PublicClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> int:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.getThrottle()
    return await function.call()


async def get_vault_proxy_address(
    client: PublicClient,
    share_price_throttled_asset_manager: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(
        share_price_throttled_asset_manager,
        abis.ISharePriceThrottledAssetManagerLib,
    )
    function = contract.functions.getVaultProxyAddress()
    return await function.call()



from typing import TypedDict
from web3.types import ChecksumAddress, HexStr, TxParams
from .utils.clients import PublicClient, WalletClient
from .configuration_lib.policy import is_enabled
from ..abis import abis


# --------------------------------------------------------------------------------------------
# DEPOSIT
# --------------------------------------------------------------------------------------------


async def get_shares_action_timelock(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> int:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getSharesActionTimelock()
    return await function.call()


async def get_last_shares_bought_timestamp(
    client: PublicClient,
    depositor: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
) -> int:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getLastSharesBoughtTimestampForAccount(depositor)
    return await function.call()


async def get_expected_shares_for_deposit(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    amount: int,
    depositor: ChecksumAddress,
) -> int:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.buyShares(amount, 1)
    return await function.call({"from": depositor})


async def deposit(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    amount: int,
    depositor: ChecksumAddress,
    min_shares_quantity: int,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.buyShares(amount, min_shares_quantity)
    return await client.populated_transaction(function, account=depositor)


# --------------------------------------------------------------------------------------------
# REDEMPTION
# --------------------------------------------------------------------------------------------


class RedeemSharesForSpecificAssetsParams(TypedDict):
    client: PublicClient
    comptroller_proxy: ChecksumAddress
    recipient: ChecksumAddress
    shares_quantity: int
    payout_assets: list[ChecksumAddress]
    payout_percentages: list[int]


async def get_specific_assets_redemption_expected_amounts(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    recipient: ChecksumAddress,
    shares_quantity: int,
    payout_assets: list[ChecksumAddress],
    payout_percentages: list[int],
) -> dict[ChecksumAddress, int]:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.redeemSharesForSpecificAssets(
        recipient, shares_quantity, payout_assets, payout_percentages
    )
    payout_amounts = await function.call({"from": recipient})
    assert all(payout_amount is not None for payout_amount in payout_amounts)
    return dict(zip(payout_assets, payout_amounts))


async def redeem_shares_for_specific_assets(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    recipient: ChecksumAddress,
    shares_quantity: int,
    payout_assets: list[ChecksumAddress],
    payout_percentages: list[int],
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.redeemSharesForSpecificAssets(
        recipient, shares_quantity, payout_assets, payout_percentages
    )
    return await client.populated_transaction(function, account=recipient)


async def redeem_shares_in_kind(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    recipient: ChecksumAddress,
    shares_quantity: int,
    additional_assets: list[ChecksumAddress],
    assets_to_skip: list[ChecksumAddress],
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.redeemSharesInKind(
        recipient, shares_quantity, additional_assets, assets_to_skip
    )
    return await client.populated_transaction(function, account=recipient)


# --------------------------------------------------------------------------------------------
# DEPOSIT WRAPPER
# --------------------------------------------------------------------------------------------


class NativeDepositArgs(TypedDict):
    client: PublicClient
    deposit_wrapper: ChecksumAddress
    comptroller_proxy: ChecksumAddress
    exchange: ChecksumAddress
    exchange_approve_target: ChecksumAddress
    exchange_data: HexStr
    min_investment_amount: int
    amount: int


async def get_expected_shares_for_native_token_deposit(
    client: PublicClient,
    deposit_wrapper: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
    exchange: ChecksumAddress,
    exchange_approve_target: ChecksumAddress,
    exchange_data: HexStr,
    min_investment_amount: int,
    amount: int,
    depositor: ChecksumAddress,
) -> int:
    contract = client.contract(deposit_wrapper, abis.IDepositWrapper)
    function = contract.functions.exchangeEthAndBuyShares(
        comptroller_proxy,
        1,
        exchange,
        exchange_approve_target,
        exchange_data,
        min_investment_amount,
    )
    return await function.call({"from": depositor, "value": amount})


async def deposit_native_token(
    client: WalletClient,
    deposit_wrapper: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
    exchange: ChecksumAddress,
    exchange_approve_target: ChecksumAddress,
    exchange_data: HexStr,
    min_investment_amount: int,
    amount: int,
    min_shares_quantity: int,
) -> TxParams:
    contract = client.contract(deposit_wrapper, abis.IDepositWrapper)
    function = contract.functions.exchangeEthAndBuyShares(
        comptroller_proxy,
        min_shares_quantity,
        exchange,
        exchange_approve_target,
        exchange_data,
        min_investment_amount,
    )
    return await client.populated_transaction(function, value=amount)


class ERC20DepositArgs(TypedDict):
    client: WalletClient
    deposit_wrapper: ChecksumAddress
    comptroller_proxy: ChecksumAddress
    input_asset: ChecksumAddress
    max_input_asset_amount: int
    exchange: ChecksumAddress
    exchange_approve_target: ChecksumAddress
    exchange_data: HexStr
    exchange_min_received: int


async def get_expected_shares_for_erc20_deposit(
    client: PublicClient,
    deposit_wrapper: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
    input_asset: ChecksumAddress,
    max_input_asset_amount: int,
    exchange: ChecksumAddress,
    exchange_approve_target: ChecksumAddress,
    exchange_data: HexStr,
    exchange_min_received: int,
    depositor: ChecksumAddress,
) -> int:
    contract = client.contract(deposit_wrapper, abis.IDepositWrapper)
    function = contract.functions.exchangeErc20AndBuyShares(
        comptroller_proxy,
        1,
        input_asset,
        max_input_asset_amount,
        exchange,
        exchange_approve_target,
        exchange_data,
        exchange_min_received,
    )
    return await function.call({"from": depositor})


async def deposit_erc20(
    client: WalletClient,
    deposit_wrapper: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
    input_asset: ChecksumAddress,
    max_input_asset_amount: int,
    exchange: ChecksumAddress,
    exchange_approve_target: ChecksumAddress,
    exchange_data: HexStr,
    exchange_min_received: int,
    min_shares_quantity: int,
) -> TxParams:
    contract = client.contract(deposit_wrapper, abis.IDepositWrapper)
    function = contract.functions.exchangeErc20AndBuyShares(
        comptroller_proxy,
        min_shares_quantity,
        input_asset,
        max_input_asset_amount,
        exchange,
        exchange_approve_target,
        exchange_data,
        exchange_min_received,
    )
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# SHARES WRAPPER DEPOSIT
# --------------------------------------------------------------------------------------------


class SharesWrapperDepositBaseParams(TypedDict):
    client: WalletClient
    shares_wrapper: ChecksumAddress
    deposit_asset: ChecksumAddress
    deposit_amount: int


async def get_expected_shares_for_shares_wrapper_deposit(
    client: PublicClient,
    shares_wrapper: ChecksumAddress,
    deposit_asset: ChecksumAddress,
    deposit_amount: int,
    depositor: ChecksumAddress,
) -> int:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.deposit(deposit_asset, deposit_amount, 1)
    return await function.call({"from": depositor})


async def shares_wrapper_deposit(
    client: WalletClient,
    shares_wrapper: ChecksumAddress,
    deposit_asset: ChecksumAddress,
    deposit_amount: int,
    min_shares_amount: int,
) -> TxParams:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.deposit(
        deposit_asset, deposit_amount, min_shares_amount
    )
    return await client.populated_transaction(function)


async def shares_wrapper_request_deposit(
    client: WalletClient,
    shares_wrapper: ChecksumAddress,
    deposit_asset: ChecksumAddress,
    deposit_amount: int,
) -> TxParams:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.requestDeposit(deposit_asset, deposit_amount)
    return await client.populated_transaction(function)


async def shares_wrapper_cancel_request_deposit(
    client: WalletClient,
    shares_wrapper: ChecksumAddress,
    deposit_asset: ChecksumAddress,
) -> TxParams:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.cancelRequestDeposit(deposit_asset)
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# SHARES WRAPPER REDEMPTION
# --------------------------------------------------------------------------------------------


async def shares_wrapper_request_redeem(
    client: WalletClient,
    shares_wrapper: ChecksumAddress,
    amount: int,
) -> TxParams:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.requestRedeem(amount)
    return await client.populated_transaction(function)


async def shares_wrapper_cancel_request_redeem(
    client: WalletClient,
    shares_wrapper: ChecksumAddress,
) -> TxParams:
    contract = client.contract(shares_wrapper, abis.IGatedRedemptionQueueSharesWrapperLib)
    function = contract.functions.cancelRequestRedeem()
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# SINGLE ASSET REDEMPTION QUEUE
# --------------------------------------------------------------------------------------------


async def redemption_queue_request_redeem(
    client: WalletClient,
    redemption_queue: ChecksumAddress,
    amount: int,
) -> TxParams:
    contract = client.contract(redemption_queue, abis.ISingleAssetRedemptionQueueLib)
    function = contract.functions.requestRedeem(amount)
    return await client.populated_transaction(function)


async def redemption_queue_withdraw_request(
    client: WalletClient,
    redemption_queue: ChecksumAddress,
    request_id: int,
) -> TxParams:
    contract = client.contract(redemption_queue, abis.ISingleAssetRedemptionQueueLib)
    function = contract.functions.withdrawRequest(request_id)
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# POLICY CHECK
# --------------------------------------------------------------------------------------------


async def is_allowed_depositor(
    client: PublicClient,
    allowed_deposit_recipients_policy: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
    policy_manager: ChecksumAddress,
    depositor: ChecksumAddress,
) -> bool:
    has_allowed_depositor_policy = await is_enabled(
        client, allowed_deposit_recipients_policy, policy_manager, comptroller_proxy
    )

    if not has_allowed_depositor_policy:
        return True

    contract = client.contract(
        allowed_deposit_recipients_policy, abis.IAllowedDepositRecipientsPolicy
    )
    function = contract.functions.passesRule(comptroller_proxy, depositor)
    return await function.call()



from ..assets_lib.ethereum import ASSETS
from ..networks import NETWORK_BY_SLUG


DEPLOYMENT = {
    "address": "0xC3DC853dD716bd5754f421ef94fdCbac3902ab32",
    "assets": ASSETS,
    "external_contracts": {
        "aaveUIIncentiveDataProvider": "0x5a40cDe2b76Da2beD545efB3ae15708eE56aAF9c",
        "aaveV2IncentivesController": "0xd784927Ff2f95ba542BfC824c8a8a98F3495f6b5",
        "aaveV2LendingPoolProvider": "0xB53C1a33016B2DC2fF3653530bfF1848a515c8c5",
        "aaveV3LendingPoolProvider": "0x2f39d218133AFaB8F2B819B1066c7E434Ad94E9e",
        "aaveV3ProtocolDataProvider": "0x41393e5e337606dc3821075Af65AeE84D7688CBD",
        "aaveV3RewardsController": "0x8164Cc65827dcFe994AB23944CBC90e0aa80bFcb",
        "aliceOrderManager": "0x841473a19279E54a850E9083A3A57dE9e6244d2E",
        "arrakisV2Helper": "0x89E4bE1F999E3a58D16096FBe405Fc2a1d7F07D6",
        "arrakisV2Resolver": "0x535C5fDf31477f799366DF6E4899a12A801cC7b8",
        "balancerGaugeController": "0xC128468b7Ce63eA702C1f104D55A2566b13D3ABD",
        "balancerHelpers": "0x5aDDCCa35b7A0D07C74063c48700C8590E87864E",
        "balancerMinter": "0x239e55F427D44C3cc793f49bFB507ebe76638a2b",
        "balancerProtocolFeesCollector": "0xce88686553686DA562CE7Cea497CE749DA109f9F",
        "balancerVault": "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
        "chainlinkFeedsRegistry": "0x47Fb2585D2C56Fe188D0E6ec628a38b74fCeeeDf",
        "compoundComptroller": "0x3d9819210A31b4961b30EF54bE2aeD79B9c9Cd3B",
        "compoundV3Rewards": "0x1B0e765F6224C21223AeA2af16c1C46E38885a40",
        "curveChildLiquidityGaugeFactory": "0x0000000000000000000000000000000000000000",
        "curveMinter": "0xd061D61a4d941c39E5453435B6345Dc261C2fcE0",
        "curveRegistry": "0x0000000022D53366457F9d5E68Ec105046FC4383",
        "cvxCrvStaking": "0x3Fe65692bfCD0e6CF84cB1E7d24108E434A7587e",
        "gmxV2ChainlinkPriceFeed": "0x0000000000000000000000000000000000000000",
        "gmxV2DataStore": "0x0000000000000000000000000000000000000000",
        "gmxV2ExchangeRouter": "0x0000000000000000000000000000000000000000",
        "gmxV2Reader": "0x0000000000000000000000000000000000000000",
        "gmxV2ReferralStorage": "0x0000000000000000000000000000000000000000",
        "kilnStaking": "0x0816DF553a89c4bFF7eBfD778A9706a989Dd3Ce3",
        "lidoWithdrawalsQueue": "0x889edC2eDab5f40e902b864aD4d7AdE8E412F9B1",
        "liquityCollSurplusPool": "0x3D32e8b97Ed5881324241Cf03b2DA5E2EBcE5521",
        "liquityHintHelpers": "0xE84251b93D9524E0d2e621Ba7dc7cb3579F997C0",
        "liquitySortedTroves": "0x8FdD3fbFEb32b28fb73555518f8b361bCeA741A6",
        "liquityTroveManager": "0xA39739EF8b0231DbFA0DcdA07d7e29faAbCf4bb2",
        "makerMCDPotAddress": "0x197E90f9FAD81970bA7976f33CbD77088E5D7cf7",
        "morphoBlue": "0xBBBBBbbBBb9cC5e90e3b3Af64bdAF62C37EEFFCb",
        "multicall": "0xcA11bde05977b3631167028862bE2a173976CA11",
        "paraswapV5AugustusSwapper": "0xDEF171Fe48CF0115B1d80b88dc8eAB59176FEe57",
        "paraswapV5TokenTransferProxy": "0x216B4B4Ba9F3e719726886d34a177484278Bfcae",
        "pendlePtLpOracle": "0x66a1096C6366b2529274dF4f5D8247827fe4CEA8",
        "staderStakingPoolManager": "0xcf5EA1b38380f6aF39068375516Daf40Ed70D299",
        "staderUserWithdrawManager": "0x9F0491B32DBce587c50c4C43AB303b06478193A7",
        "stakeWiseV3KeeperRewards": "0x6B5815467da09DaA7DC83Db21c9239d98Bb487b5",
        "theGraphDelegationStakingProxy": "0xF55041E37E12cD407ad00CE2910B8269B01263b9",
        "theGraphEpochManagerProxy": "0x64F990Bf16552A693dCB043BB7bf3866c5E05DdB",
        "uniswapV3NonFungiblePositionManager": "0xC36442b4a4522E871399CD717aBDD847Ab11FE88",
        "voteLockedConvexToken": "0x72a19342e8F1838460eBFCCEf09F6585e32db86E",
        "votiumVoteProxy": "0xde1E6A7ED0ad3F61D531a8a78E83CcDdbd6E0c49",
        "zeroExExchangeProxy": "0x95E6F48254609A6ee006F7D493c8e5fB97094ceF",
        "zeroExV4Exchange": "0xDef1C0ded9bec7F1a1670819833240f027b25EfF",
        "zeroLendAaveV3UIIncentiveDataProvider": "0x0A1198DDb5247a283F76077Bb1E45e5858ee100b",
        "zeroLendLRTBTCAaveV3LendingPoolProvider": "0x17878AFdD5772F4Ec93c265Ac7Ad8E2b29abB857",
        "zeroLendLRTBTCAaveV3ProtocolDataProvider": "0x31063F7CA8ef4089Db0dEdf8D6e35690B468A611",
        "zeroLendLRTBTCAaveV3RewardsController": "0x938e23c10C501CE5D42Bc516eCFDf5AbD9C51d2b",
        "zeroLendRWAStablecoinsAaveV3LendingPoolProvider": "0xE3C3c5eAd58FC2BEd4E577E38985B8F7F1DDfF00",
        "zeroLendRWAStablecoinsAaveV3ProtocolDataProvider": "0x298ECDcb0369Aef75cBbdA3e46a224Cfe622E287",
        "zeroLendRWAStablecoinsAaveV3RewardsController": "0xbc2FDC58E3E06D265B46E32771442b7f5098725b",
    },
    "inception": 11636493,
    "kind": "live",
    "known_address_lists": {
        "no_slippage_adapters": 1,
        "adapters": 2,
        "fees": 3,
        "policies": 4,
        "kiln_staking_contracts": 400,
        "non_standard_price_feed_assets": 650,
        "a_tokens": 463,
        "zero_lend_rwa_stablecoins_a_tokens": 737,
        "zero_lend_lrt_btc_a_tokens": 738,
        "deposit_wrapper_allowed_exchanges": 553,
    },
    "known_uint_lists": {
        "allowed_morpho_blue_vaults": 3,
    },
    "label": "Ethereum",
    "named_tokens": {
        "aave": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
        "bal": "0xba100000625a3754423978a60c9317c58a424e3D",
        "ceth": "0x4Ddc2D193948926D02f9B1fE9e1daa0718270ED5",
        "comp": "0xc00e94Cb662C3520282E6f5717214004A7f26888",
        "crv": "0xD533a949740bb3306d119CC777fa900bA034cd52",
        "cvx": "0x4e3FBD56CD56c3e72c1403e103b45Db9da5B9D2B",
        "dai": "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "diva": "0xBFAbdE619ed5C4311811cF422562709710DB587d",
        "ethx": "0xA35b1B31Ce002FBF2058D22F30f95D405200A15b",
        "grt": "0xc944E90C64B2c07662A292be6244BDf05Cda44a7",
        "idle": "0x875773784Af8135eA0ef43b5a374AaD105c5D39e",
        "lusd": "0x5f98805A4E8be255a32880FDeC7F6728C6568bA0",
        "mln": "0xec67005c4E498Ec7f55E092bd1d35cbC47C91892",
        "mpl": "0x33349B282065b0284d756F0577FB39c158F935e6",
        "paxg": "0x45804880De22913dAFE09f4980848ECE6EcbAf78",
        "ptkn_mln": "0x4C0d9e35AD2C952bCd58A6d255C9Dc2784AC6f08",
        "sthoundeth": "0xDFe66B14D37C77F4E9b180cEb433d1b164f0281D",
        "stkaave": "0x4da27a545c0c5B758a6BA100e3a049001de870f5",
        "steth": "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84",
        "stusd": "0x0022228a2cc5E7eF0274A7Baa600d44da5aB5776",
        "sweth": "0xf951E335afb289353dc249e82926178EaC7DEd78",
        "uni": "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
        "usda": "0x0000206329b97DB379d5E1Bf586BbDB969C63274",
        "usdc": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
        "usdt": "0xdAC17F958D2ee523a2206206994597C13D831ec7",
        "weth": "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
    },
    "network": NETWORK_BY_SLUG["ethereum"],
    "releases": {
        "sulu": {
            "address": "0x4f1C53F096533C04d8157EFB6Bca3eb22ddC6360",
            "contracts": {
                "AaveDebtPositionLib": "0xb3402eb0d45c63594699DF3E5caC9116Ccfc5eF3",
                "AaveDebtPositionParser": "0xf6d47bdEe7760067557B047cE4AdCba819d7D5A8",
                "AavePriceFeed": "0xE9cFe0f99b8a01Fd80F110Da4d8F08f6bF3DD6A6",
                "AaveV2Adapter": "0xECe6B376af7C9273cebaf6528565c47Ea2Cb8a4C",
                "AaveV2ATokenListOwner": "0x26Cd328f2bdFe13A99b7e2FC19BbdCB4d8B54D69",
                "AaveV3Adapter": "0x9cfb64D91Ce4eB821fF8EdC1C2fdA2E89E256707",
                "AaveV3ATokenListOwner": "0xCFe249a7AE4619980eeA1A2d83a26a5E7281EbB0",
                "AaveV3DebtPositionLib": "0x61a063484d5c124B23740aa27688e2b0CCb562d5",
                "AaveV3DebtPositionParser": "0x81C2a0AC95549FF0F0Fa2d8c3136cc8D4b2439E3",
                "AaveV3FlashLoanAssetManagerFactory": "0xD6e8B30214eF909421EB1d20E5B281777810004A",
                "AaveV3FlashLoanAssetManagerLib": "0x8Aa6F11fEF40Eb24c2f77de538F40aC8a9Ec9f0d",
                "AddressListRegistry": "0x4eb4c7Babfb5d54ab4857265B482Fb6512d22DFF",
                "AlicePositionLib": "0xE57c868F97E8c1DEAF24DC72b89D886EBFa24D9f",
                "AlicePositionParser": "0xFF60eB4b559d63bEbA1738281B63704266C6dAc0",
                "AllowedAdapterIncomingAssetsPolicy": "0x2f0E55830a173d845A886FD574F01a039A07Fc37",
                "AllowedAdaptersPerManagerPolicy": "0xA4507d51c5270Ff91229b76300FF90774384d144",
                "AllowedAdaptersPolicy": "0x720ef97bF835699FCF07591952cD2B132D63a6C0",
                "AllowedAssetsForRedemptionPolicy": "0x823ca839DA344DA59D517b84ce3baB9FFc9f54eE",
                "AllowedDepositRecipientsPolicy": "0xA66BAAa0cCb6468C5A2cB61f5D672C7bA0440Ee1",
                "AllowedExternalPositionTypesPerManagerPolicy": "0x47fB78995D945D501f6f9bAd343d7Ce7d3dB54AB",
                "AllowedExternalPositionTypesPolicy": "0x9e076e7d35a3b881Ab9e3DA958431630FdFa756F",
                "AllowedRedeemersForSpecificAssetsPolicy": "0xEd4AA74490843E7ad64B445eF18CecE8A0562433",
                "AllowedSharesTransferRecipientsPolicy": "0xebE37e43bC6b3AacfE318d6906fc80C4a2a7505A",
                "ArbitraryLoanPositionLib": "0x73CB96137cb5455E77275A6aB3411d0d52d545A9",
                "ArbitraryLoanPositionParser": "0xE959C0eeF487f7eE098aD10998d9dfCF4fa1d1aF",
                "ArbitraryLoanTotalNominalDeltaOracleModule": "0x7F2A48122BBd3ffba33Ed9D1f5cFAbEdE7Caab34",
                "ArrakisV2Adapter": "0xC6e54fb0f526D3516224a5E09BD435323800ad94",
                "ArrakisV2PriceFeed": "0xffF8D5C955D9d71f3A09E166521e12e28d3042c7",
                "AssetValueCalculator": "0xFa3D84D841C55bb9175b0a6f215E2300c40CC5Ab",
                "BalancerV2GaugeTokenPriceFeed": "0x312Ce4bDE393Ff246b27c890b97ADe37fEdaA348",
                "BalancerV2LiquidityAdapter": "0xAB5dA4aa08B56c7e5A9D5d8A5fF19cF09a88c305",
                "BalancerV2StablePoolPriceFeed": "0x438254D66e2bC576819A76A0ecB11FE41195D39F",
                "BalancerV2WeightedPoolPriceFeed": "0x083B8F465BbA2BB274E48387e3f9C56323341286",
                "ChainlinkLikeWstethPriceFeed": "0x92829C41115311cA43D5c9f722f0E9e7b9fcd30a",
                "ChainlinkLikeYnEthPriceFeed": "0xa8f6033cE40FaB29C228f32Ef44d38cB3043c5Bc",
                "CompoundAdapter": "0x6ce8095a692afF6698c3Aa8593BE3976B6b8743D",
                "CompoundDebtPositionLib": "0xc8e0c169dAE4cdfD2F51E7494D2B5459F9314ee9",
                "CompoundDebtPositionParser": "0x8BFbd013020c4D26D043183eff600ea3b36B4afb",
                "CompoundPriceFeed": "0x6254e0538Abbb668eaC89d5e5BEe27a9D0E62bBE",
                "CompoundV3TokenListOwner": "0xa9E085B1CA8ECd1dd5BF4c1B5feD3BE84C341Fd4",
                "CompoundV3Adapter": "0xFaA9B9Cc98503F51A54F6038DfdD0E43AA0Ac98e",
                "ComptrollerLib": "0x03F7f3B8Da875881206655D8099B9DACf721f1EF",
                "ConvertedQuoteAggregatorFactory": "0x8F1E6f61323Bf64A47fFbfCa99e9e2Bd4f982c07",
                "ConvexVotingPositionLib": "0x12Fa6805a1ff2d21318Dcbcf677712Bde8a033E1",
                "ConvexVotingPositionParser": "0x4619b0394f09Ef964407deDCe4ca19AD012BCa20",
                "CumulativeSlippageTolerancePolicy": "0x3A49D5aEC385Ac1bDe99F305316b945C5Ee71312",
                "CurveExchangeAdapter": "0xC9c14a99cCF467EA1ff2e19584A5FaBA3671B8de",
                "CurveLiquidityAaveAdapter": "0x83E249d603459D2A926EE85d8ba8d0AeABe101F2",
                "CurveLiquidityAdapter": "0xEa0a896ddE31CFcB53A96ac767119B69D7B4f633",
                "CurveLiquiditySethAdapter": "0x87Db8e815f9BDE3b5b686e3E892FAF0E86481c44",
                "CurveLiquidityStethAdapter": "0x1453E5a5C029cdd91f584c603818baf2ab4CE663",
                "CurvePriceFeed": "0x7e6130e497035a34A08E0F18B7A9B0fa06ed013E",
                "DepositWrapper": "0x65BbAD6545B7ac9C30Fb0f07e64e25106Bf05eEc",
                "DisallowedAdapterIncomingAssetsPolicy": "0x5E216f370e3555FEb9A0575A57AdA732A9e50386",
                "Dispatcher": "0xC3DC853dD716bd5754f421ef94fdCbac3902ab32",
                "EntranceRateBurnFee": "0xCDEc5BBECC6d2C004D5378A63a3c484c2643eD9d",
                "EntranceRateDirectFee": "0xFb8DF7D5e320020Cd8047226b81cf6d68F3E3C19",
                "ERC4626Adapter": "0x64Fa106DD89F21d6e687EEbE9384637F7d54f707",
                "ERC4626PriceFeed": "0x66aA5b2FdFB453F8A27f9BD1d9124947Ef3886BB",
                "EtherFiEthPriceFeed": "0x31103dB5639AD1d5351B83409fDf7E575e26774b",
                "ExitRateBurnFee": "0x06b13918E988D1314dA1a9dA4C0cdE5fe994364a",
                "ExitRateDirectFee": "0x3a09d11C20AA1aD38C77B4f426901d3427f73fBE",
                "ExternalPositionFactory": "0x0AAcb782205dde9eFf4862ACe9849DcE1ca3409f",
                "ExternalPositionManager": "0x1e3dA40f999Cf47091F869EbAc477d84b0827Cf4",
                "FeeManager": "0xAf0DFFAC1CE85c3fCe4c2BF50073251F615EefC4",
                "FiduPriceFeed": "0x478335491E1fB798c64a9E95CA43d3D967F42486",
                "FundDataProviderRouter": "0x781ec8517a75172EA007d7458ea1811fC9F99C30",
                "FundDeployer": "0x4f1C53F096533C04d8157EFB6Bca3eb22ddC6360",
                "FundValueCalculator": "0x490e64E0690b4aa481Fb02255aED3d052Bad7BF1",
                "FundValueCalculatorRouter": "0x7c728cd0CfA92401E01A4849a01b57EE53F5b2b9",
                "GasRelayPaymasterFactory": "0x846bbe1925047023651de7EC289f329c24ded3a8",
                "GasRelayPaymasterLib": "0x2F1D153DD821258Caa15656BDe910e146e5F2f77",
                "GatedRedemptionQueueSharesWrapperFactory": "0x73b9C40530311b49B526f230D01Bdf5725B3290d",
                "GatedRedemptionQueueSharesWrapperLib": "0xE971375e3E8af54232F9B7c88cCE143EDf95C272",
                "GenericAdapter": "0x0000000000000000000000000000000000000000",
                "GlobalConfigLib": "0x6682E70860d48a039F52DaCCdA917250349a3fB3",
                "GlobalConfigProxy": "0x5611dF74A77EfD198De5Fc7f83A482DcFE0c7A7A",
                "GMXV2LeverageTradingPositionLib": "0x0000000000000000000000000000000000000000",
                "GMXV2LeverageTradingPositionParser": "0x0000000000000000000000000000000000000000",
                "IntegrationManager": "0x31329024f1a3E4a4B3336E0b1DfA74CC3FEc633e",
                "KilnStakingPositionLib": "0x266c98eF7C16c791a27Dc60E8877e8e3954090A2",
                "KilnStakingPositionParser": "0xFa9a7f8f30c9BC741BE5959af8Cd48636A6aecC2",
                "LidoWithdrawalsPositionLib": "0xa511ecEa62F360E983829e408Fc753ADcADcdEAE",
                "LidoWithdrawalsPositionParser": "0x501083bE98eBb2d75bE75459aCcc5C5922C07F28",
                "LiquityDebtPositionLib": "0x016b63941Df5fD92B49F861559d8B7792629419F",
                "LiquityDebtPositionParser": "0x55e41333D949F13C73aE99689d7137574D927d4B",
                "ManagementFee": "0xFaF2c3DB614E9d38fE05EDc634848BE7Ff0542B9",
                "ManualValueOracleFactory": "0x0edbb060A8f00F5967eECFC87c8559fA65501A3D",
                "MapleLiquidityPositionLib": "0xB35623A479383b87da32fdDe8dBe83FaC5f6bf3D",
                "MapleLiquidityPositionParser": "0x31915278772A2bC68cBFb7Effe1209677b0Aad10",
                "MinAssetBalancesPostRedemptionPolicy": "0x58C0a2A546b3903fA68a53e34EE0C8A02AABFAD0",
                "MinMaxInvestmentPolicy": "0xebdadFC929c357d12281118828AeA556db5be30C",
                "MinSharesSupplyFee": "0xbC9dA8eddE80ffb1294852d23EE1b385Ea2d4929",
                "MorphoBluePositionLib": "0xd84123F688601133d478949277BE8065f48Ff810",
                "MorphoBluePositionParser": "0xfDf34798B21774523Be370490b8730807d4f9167",
                "NoDepegOnRedeemSharesForSpecificAssetsPolicy": "0x96ef0f7c10505460fA39C57a037a5EC2520b8B25",
                "NotionalV2PositionLib": "0x73436d870dA10281Aa3872b1eC48451e077900DE",
                "NotionalV2PositionParser": "0x59053fC5ee387321ea4aB415d1160e49E4f5B166",
                "OneInchV5Adapter": "0x6C62b8F7b2fd1c60fFD3Afc1A2B15d4318745677",
                "OnlyRemoveDustExternalPositionPolicy": "0x966eC191ED9e026Cb6f7e22BB2A284bad6A2838d",
                "OnlyUntrackDustOrPricelessAssetsPolicy": "0x747bEAEe139FbA4a89fA71bEBB5F21231530292B",
                "ParaSwapV5Adapter": "0x871a7f0eF4917a1534E651d1fDe3763A52A23ECe",
                "PeggedDerivativesPriceFeed": "0x86533352BDd201c89f184f7ebbFeBea3E31c8Bb3",
                "PerformanceFee": "0xfeDC73464Dfd156d30F6524654a5d56E766DA0c3",
                "PendleV2Adapter": "0xe77Ba2e88Aae1543839eC6Ee0A0f847391205610",
                "PendleV2PositionLib": "0x85C474A1dbf7fDc22C592382aD554Bb42CebcE0b",
                "PendleV2PositionParser": "0x559AF4E5765D60003E4895d23080C501CBB9ED3C",
                "PendleMarketsRegistry": "0x31391adF6402fDd4eb4b57801A612b0987a0B0f2",
                "PolicyManager": "0xADF5A8DB090627b153Ef0c5726ccfdc1c7aED7bd",
                "PoolTogetherV4Adapter": "0x0000000000000000000000000000000000000000",
                "PoolTogetherV4PriceFeed": "0x0000000000000000000000000000000000000000",
                "ProtocolFeeReserveLib": "0xA0ed89af63367Ddc8e1dd6b992F20D1214cCB51C",
                "ProtocolFeeReserveProxy": "0xB7460593BD222E24a2bF4393aa6416bD373995E0",
                "ProtocolFeeTracker": "0xe97980f1D43C4CD4F1EeF0277a2DeA7ddBc2Cd13",
                "SharePriceThrottledAssetManagerFactory": "0x0883BA10F44217B97bDE11900e197738a7dF911B",
                "SharePriceThrottledAssetManagerLib": "0xFdE8c198BeF60D026332a671F64c34D65C60C935",
                "SharesSplitterFactory": "0xfc8ed755c52782fA1A4bA9193B566E775701E511",
                "SingleAssetRedemptionQueueFactory": "0xfE84D5209054254389C9D6a754B821f3a297D56a",
                "SingleAssetRedemptionQueueLib": "0x0012B7C26b8C081a29A61cD52526cF6305367968",
                "SolvV2BondBuyerPositionLib": "0x0E7D828f9F9A3Ce152D39f21F2e2D0fF89448b6b",
                "SolvV2BondBuyerPositionParser": "0xA2Ed786215d6F4DA95338ee7abB84f28134dc19c",
                "SolvV2BondIssuerPositionLib": "0x31c30bAcb054f5dB7f5bC96850b488440AbB8991",
                "SolvV2BondIssuerPositionParser": "0xC4B599043a5479398eb8Af387b1E36D9A924F8C2",
                "StaderSDPriceFeed": "0x9938B14A25a4910531d5cBdf3c41510b19aaF016",
                "StaderStakingAdapter": "0x7F1B68d5Ed183CdA6788A66520506eAF3544001C",
                "StaderWithdrawalsPositionLib": "0x5Cf43F5f8c1648Db23948e3814d0099c408201a4",
                "StaderWithdrawalsPositionParser": "0xaeFE3260dCBCfaA2A4b927A6494057837E6dD902",
                "StakeWiseV3StakingPositionLib": "0xeF268Eb475a096AdB95712b05b095aCc1Fc2d3aE",
                "StakeWiseV3StakingPositionParser": "0x642348Ee0c28C2943082f950F3A35dB07Bd4bbDF",
                "SwellStakingAdapter": "0x50bAe03333Dd8495263c9049091a8925063b068E",
                "SynthetixAdapter": "0x4bAF8282571FEbdfd8768b475551EC14e144Edb3",
                "TermFinanceV1LendingPositionLib": "0x2817AB33b56ec310621592BfDbE1fAd6470D19a0",
                "TermFinanceV1LendingPositionParser": "0xa95C7Bab91692df52255b8Ad50C5eD2bcC9064aD",
                "TheGraphDelegationPositionLib": "0x6fFf66d55698A601e91989c44349Da8a2A9A7848",
                "TheGraphDelegationPositionParser": "0x2226D7687109d6B6A0882f8EEF2b4A4c90dc677E",
                "ThreeOneThirdAdapter": "0x5a1c0E89133C4Cd844A8B345370565f1368A79A8",
                "TransferAssetsAdapter": "0xe0309Fa2412B811a0BD40A73297093707259217f",
                "UintListRegistry": "0x6FfD6fC068E7b365AF18dA4fdC39D3289159407B",
                "UniswapV2ExchangeAdapter": "0x8c36435A653041BFd65515CC82502663C1ce6F0e",
                "UniswapV2LiquidityAdapter": "0xF78130AfedA6d9Df3394b34d36239aeC7FAe48d9",
                "UniswapV2PoolPriceFeed": "0xdbc1162eA1aB770f0BA5Fb494e0010e68351e3b0",
                "UniswapV3Adapter": "0xeD6A08E05cB4260388dc7CC60Bc5fEFcCfab2793",
                "UniswapV3LiquidityPositionLib": "0x20A2D4765bE139475C34DB7B7d856DCF25092C26",
                "UniswapV3LiquidityPositionParser": "0x23805FED4b73a7b77c28f2823733736951C49d6C",
                "UnpermissionedActionsWrapper": "0xCFAB4FCBFE059d5C1840D9dc285a9BFA0F96A118",
                "UsdEthSimulatedAggregator": "0x9579f735d0C93B5eef064Fe312CA3509BD695206",
                "ValueInterpreter": "0xD7B0610dB501b15Bfb9B7DDad8b3869de262a327",
                "VaultLib": "0x891dee0483eBAA922E274ddD2eBBaA2D33468A38",
                "WstethPriceFeed": "0x50Da4957032c8fC5F94eC8D5EC8bFcE84f9C9311",
                "YearnVaultV2Adapter": "0x7eA777F9F6ecBf4d03Dc5323d3F057B0730Fc34a",
                "YearnVaultV2PriceFeed": "0x3982e1CC26b99310747Df54f445063745C54A324",
                "ZeroExV2Adapter": "0xace8e944cDa48439E8eeDa4027115CFb6d942854",
                "ZeroExV4Adapter": "0xe073406C5306043Eda44F09C67654A7729277a47",
                "ZeroExV4AdapterPmm2Kyc": "0x1376a75316B6Ec7BfFaA0660f34D852A2c0e307B",
                "ZeroLendLRTBTCAaveV3Adapter": "0xbc91Fe394d7d372fbae757dbffd32B44A29b3F6E",
                "ZeroLendLRTBTCAaveV3ATokenListOwner": "0x7AFfC39D73CB7d22a7FD099a16441D9eb96eF31B",
                "ZeroLendLRTBTCAaveV3DebtPositionLib": "0x0958DCa5eF97F1aB0b931bA6395Ef40625c26C9E",
                "ZeroLendLRTBTCAaveV3DebtPositionParser": "0xC4701ABc4E1Ae1ac028f98C0fa9395603bd079E1",
                "ZeroLendRWAStablecoinsAaveV3Adapter": "0x8525B3b6179690B74d7F067F131f780a3Ef1562C",
                "ZeroLendRWAStablecoinsAaveV3ATokenListOwner": "0x55C3498Ee0D5c79b034554f301d8a799bAE7D278",
                "ZeroLendRWAStablecoinsAaveV3DebtPositionLib": "0x81936cd39aa737Feed06Ee6E7EF8aC49c63B99c6",
                "ZeroLendRWAStablecoinsAaveV3DebtPositionParser": "0x05D10E09B4FbcA7e1Dac9bF1FdA804Acc19Ec4c7",
            },
            "inception": 14132862,
            "network": NETWORK_BY_SLUG["ethereum"],
            "slug": "ethereum.sulu",
            "status": "live",
            "version": "sulu",
        },
        "encore": {
            "address": "0x7e6d3b1161DF9c9c7527F68d651B297d2Fdb820B",
            "contracts": {
                "AaveAdapter": "0x32d0CB26C2964749bd5D10Ba93695e52b4B57eA3",
                "AavePriceFeed": "0x4e49a272dC42E26C8772de366aaa93d1AC816794",
                "AdapterBlacklist": "0xA2A031B137452986B0E8fA32c9430D29b9b0494F",
                "AdapterWhitelist": "0x306f571110C5441a8481d3CD7E8B16A7E2E967E3",
                "AggregatedDerivativePriceFeed": "0x2E45f9B3fd5871cCaf4eB415dFCcbDD126F57C4f",
                "AlphaHomoraV1Adapter": "0x63765Ae5Df9bB8AF9D8bc976E595b79907CF341c",
                "AlphaHomoraV1PriceFeed": "0x1a8b4f6D469Ee775d5982C4d4AAb46677c2C92d4",
                "AssetBlacklist": "0xdC1e40174ad831e505E8191881A66e90c3681E33",
                "AssetValueCalculator": "0xa85cbA5B7a1139848C5b8C6dfFc3E4107bF8f018",
                "AssetWhitelist": "0xA3b7B872407452E38dFfACa8dA2dB3eD6D756Ad1",
                "BuySharesCallerWhitelist": "0xBc91dcc90cCdd0B51e3C623528cb886358C60105",
                "ChainlinkPriceFeed": "0x1fad8faf11E027f8630F394599830dBeb97004EE",
                "CompoundAdapter": "0xb7bfd3582582A82Ef2A2f0Df0adf705f35A07D4B",
                "CompoundPriceFeed": "0x6d7F71Cc3109D4132Ad6124D84e72E353b979880",
                "ComptrollerLib": "0x1A7CBEDf13E25818dC377d2B1277E03e5d496300",
                "CurveExchangeAdapter": "0x1990D8Bc382D38fA21EE3B570e9d1E6272152165",
                "CurveLiquidityAaveAdapter": "0x78bA80C8517A4f3146FD76A3642Fa57B77F331eb",
                "CurveLiquidityEursAdapter": "0x893b4FA60eFBABFC35a54795E2ea9Eb97344f64D",
                "CurveLiquiditySethAdapter": "0xeE91E3bB4752F358DDAaC6ab825536698b239546",
                "CurveLiquidityStethAdapter": "0x6367bb496390dDf0ef508Db2992F53DBf376b1f4",
                "CurvePriceFeed": "0xc106f1B01017c854a9cd2D88Db733408236DD809",
                "Dispatcher": "0xC3DC853dD716bd5754f421ef94fdCbac3902ab32",
                "EntranceRateBurnFee": "0x27F74B89B29508091014487253d8D9b88aa0264A",
                "EntranceRateDirectFee": "0x30a398Eb63B62bF2865F90e37752c8300Ef22F05",
                "FeeManager": "0x5D0A363E9a17fb839e2843ffa93C808CdafCCb67",
                "FundActionsWrapper": "0xbc9a63cad5a825bD5854Abca4f3D42d6aCF9Bffa",
                "FundDataProviderRouter": "0x781ec8517a75172EA007d7458ea1811fC9F99C30",
                "FundDeployer": "0x7e6d3b1161DF9c9c7527F68d651B297d2Fdb820B",
                "FundValueCalculator": "0x21560C8c72BEBe1a240d442813F2C35B05cAc5EC",
                "FundValueCalculatorRouter": "0x7c728cd0CfA92401E01A4849a01b57EE53F5b2b9",
                "GuaranteedRedemption": "0x08d4225f0995D770C4C8ea85380602B3499164a9",
                "IdleAdapter": "0x649B6cc835C8fE3D3b94b6a9c155f791bF500DFE",
                "IdlePriceFeed": "0x13c2263e534BD27149d96b8Cb9961ea1beB560Ef",
                "IntegrationManager": "0x965ca477106476B4600562a2eBe13536581883A6",
                "InvestorWhitelist": "0x775f8D5d016C4D8a1427982148ec2170282c7784",
                "KyberAdapter": "0x4ff16EfF3b6D2175A513EF4c5b95f5f4D2F05179",
                "LidoStethPriceFeed": "0x11D8d414724281BD702838BA16C8F15F7C473e9a",
                "ManagementFee": "0x45F1b268CC6412B454Dae20F8971fc4A36aF0D29",
                "MaxConcentration": "0xDee63B253432a443A09bF8461f20256d184319de",
                "MinMaxInvestment": "0x016a7287F0fDbDcE5f903334f574b2238Be3fA25",
                "ParaSwapV4Adapter": "0x8eA6Ca02274E1b05b70c11058213ADC667258C3D",
                "PerformanceFee": "0x3c3f9143A15042B69eB314Cac211688A4E71a087",
                "PolicyManager": "0x0Bd9f0465d21d4c300c7B8d781A013bdc87A31E8",
                "StakehoundEthPriceFeed": "0xd3C515EAD7Bcc7451cd920e066C4a1849B827dfD",
                "SynthetixAdapter": "0x07698C3fA26a62EC6320872bF2950bA160A2dd7C",
                "SynthetixPriceFeed": "0x4158F02ac2fF0f0E7959A768D06619B685aBF0C8",
                "TrackedAssetsAdapter": "0xeA05F4a383adCE7F1E9cC1b6029B51d7260E53aa",
                "UniswapV2Adapter": "0x8481150a0f36c98EE859e6C7B46d61FDD836768f",
                "UniswapV2PoolPriceFeed": "0x37FFB1842eF012d83A6FFdc92A6353044Ae0e3e8",
                "ValueInterpreter": "0x10a5624840Ac07287f756777DF1DEC34d2C2d654",
                "VaultLib": "0x1b3694907Ed7459c7B0116e7c6a4B7788288577F",
                "WdgldPriceFeed": "0x76b548553d4E729c81e8ae8Ab826795bd88BA9E7",
                "YearnVaultV2Adapter": "0xDA27D95Bd940cAc017f9484175E5512D3e4B0662",
                "YearnVaultV2PriceFeed": "0x7B768DD670B6f0d4AFC270cb413A8F7ca1889BD7",
                "ZeroExV2Adapter": "0x9b6cD2195d3c275875cB469A59dd437995712550",
            },
            "inception": 12387778,
            "network": NETWORK_BY_SLUG["ethereum"],
            "slug": "ethereum.encore",
            "status": "deprecated",
            "version": "encore",
        },
        "phoenix": {
            "address": "0x9134C9975244b46692Ad9A7Da36DBa8734Ec6DA3",
            "contracts": {
                "AaveAdapter": "0x14c6b99AfFC61e9b0753146F3437A223d0c58279",
                "AavePriceFeed": "0x3996D53013f562FD6D5B777Ab0Ee681802698e0D",
                "AdapterBlacklist": "0x9f471f573414d9cAcFe3A0127aD12F195504de7B",
                "AdapterWhitelist": "0xb4dF8b4f3eC62619E8D2aed5df360aa393BB8745",
                "AggregatedDerivativePriceFeed": "0xCFb6F4C08856986d13839B1907b5c645EE95388F",
                "AlphaHomoraV1Adapter": "0xed6Bdfa2725da6687bC89Bcb0BDB1e5dEAe15838",
                "AlphaHomoraV1PriceFeed": "0x5f2E2E7FfbcAeE69d570D531A00228322632231E",
                "AssetBlacklist": "0x2A7CD39427831045e7050723006697fD0801B184",
                "AssetValueCalculator": "0x929186c01e51D00209cf520BaC0651DCA4b6Fad1",
                "AssetWhitelist": "0x664F6C88CA8d5888b11B97450A5623003b8981B9",
                "BuySharesCallerWhitelist": "0x21fc97450c65dAfe14255A06Cf9c5C88483bA309",
                "ChaiAdapter": "0xB3646e9B6dF463a2dC36de67FFD28cbAb3f8A5cC",
                "ChaiPriceFeed": "0x8bA6f2D9C81Db32665df6a926227dD00cacA7d9B",
                "ChainlinkPriceFeed": "0xfB3A9655F5feA18caC92021E83550F829ae6F7F7",
                "CompoundAdapter": "0xD050DC9E75f24Ae653d282D0cFb772871729e710",
                "CompoundPriceFeed": "0xC80bD25cd46e49277CcB56E751704A6316Af30aD",
                "ComptrollerLib": "0x94F262802806BE3646612D0705802710dD5B58dF",
                "CurveExchangeAdapter": "0x402A81aD2972a017B4564453E69afaE2b006A7E2",
                "CurveLiquidityAaveAdapter": "0xF6F7cc9464Cb15e0A5B116D738dCa88434bbA00A",
                "CurveLiquiditySethAdapter": "0xbD81a0eF03F4eD89F73a30eb33137787BF92e177",
                "CurveLiquidityStethAdapter": "0x3Bb38E67033e24E604e53964A99D914f8B216521",
                "CurvePriceFeed": "0x884D909EA86EC61128a4cFE1B455d5B440cdE9C8",
                "Dispatcher": "0xC3DC853dD716bd5754f421ef94fdCbac3902ab32",
                "EntranceRateBurnFee": "0xA831a43f99af957bc27D03963958Fd710DC1f50E",
                "EntranceRateDirectFee": "0x42bc95f119Fb08b9FC72262D255016fa5546caa4",
                "FeeManager": "0xEcDbcdB8Dbf0AC54f47E41D3DD0C7DaE07828aAa",
                "FundActionsWrapper": "0x38c9f2870003d47f704e317C10f93Ca1ddAE67C1",
                "FundDataProviderRouter": "0x781ec8517a75172EA007d7458ea1811fC9F99C30",
                "FundDeployer": "0x9134C9975244b46692Ad9A7Da36DBa8734Ec6DA3",
                "FundValueCalculator": "0xc60642c271cd1102e922027aF81e4ed4736D48C7",
                "FundValueCalculatorRouter": "0x7c728cd0CfA92401E01A4849a01b57EE53F5b2b9",
                "GuaranteedRedemption": "0xda9d4Ee6C846E0b504979ea3d607c2e98574A09C",
                "IdleAdapter": "0x2f7918d9A98c199173fb8ca989B408B4Fb1EA441",
                "IdlePriceFeed": "0x08eCd9e1378DC0262df826C07D3D64c728a46829",
                "IntegrationManager": "0xFc3f356217120318Bd46c879b3A55C3135473752",
                "InvestorWhitelist": "0x3C8a4DaCe64B238C8307E3AF563fCf76Cf870d57",
                "KyberAdapter": "0x963E9C15a26d74085ABfC8B48280b650426F998e",
                "LidoStethPriceFeed": "0x4F3Da46BF999C6b3005C1465672009fBF99e7b9f",
                "ManagementFee": "0x889f2FCB6c12d836cB8f7567A1bdfa512FE9f647",
                "MaxConcentration": "0x6b9a7604faB14721B51A89Cd1CcEF639Da8664eE",
                "MinMaxInvestment": "0xDCCae078656f20B0851Dd35683315be56AEEe8f3",
                "ParaSwapV4Adapter": "0x1C1B94812Faf3dE496a2634B36C1FbC6e5c222e7",
                "PerformanceFee": "0x70478df01108Cb2fCB23463814e648363CE17720",
                "PolicyManager": "0x4c2c07b15b0b32Bad989d9DeFaeC775e2aA8A7AD",
                "StakehoundEthPriceFeed": "0xE2B19FA2D662F4eEA51D394b71Ce7281214840C5",
                "SynthetixAdapter": "0xBF3411973F09c32E8c4e67d1345DdDC5F9c7535a",
                "SynthetixPriceFeed": "0x2180341339010C237FF8e3dd1fF24063FB18C4D0",
                "TrackedAssetsAdapter": "0xde4A42052db0FB0E8935C9F3015b8cd56A3dDF43",
                "UniswapV2Adapter": "0x972318a0f4935c3153a2AA4c81274DfF621B360b",
                "UniswapV2PoolPriceFeed": "0x9177A3354EE50bffBcC42C4c6baC27ed63979097",
                "ValueInterpreter": "0x6618bF56E1C7b6c8310Bfe4096013bEd8F191628",
                "VaultLib": "0xAC3fe07F51C51153E181bE892e4e37326EEA13Da",
                "WdgldPriceFeed": "0x45919dcDECcA3C7a0D29f959E6b78a605a17A3A2",
                "ZeroExV2Adapter": "0x4205073e7AD2F9896F827DBCe496dd2306DB602e",
            },
            "inception": 11636497,
            "network": NETWORK_BY_SLUG["ethereum"],
            "slug": "ethereum.phoenix",
            "status": "deprecated",
            "version": "phoenix",
        },
    },
    "slug": "ethereum",
    "subgraphs": {
        "assets": {
            "slug": "asset-universe",
            "id": "4ZW3mDNgpDVy68RipQLJxvRw1FReJTfvA7nbB52J4Gjg",
        },
        "balances": {
            "slug": "vault-balances",
            "id": "HwR7jTExHWNvQetTxRYEMQ5hywHyUkierAYvnGS7pBUS",
        },
        "core": {
            "slug": "enzyme-core",
            "id": "9DLBBLep5UyU16kUQRvxBCMqko4q9XzuE4XsMMpARhKK",
            "dev_version": "version/latest",
        },
        "shares": {
            "slug": "vault-shares",
            "id": "6p2L2gQ4Hw4Dh2kxZFDJbcqtbv44vrJbrBEh3EjS7qVo",
        },
        "vaults": {
            "slug": "vault-lineage",
            "id": "5FdivFcUPmVSqCFkv3jqJh3QYjHjh1ztzd7GHiCAMP1h",
        },
    },
}



from typing import TypedDict, Any, Literal, get_args
from web3 import Web3
from web3.types import ChecksumAddress
from web3.constants import ADDRESS_ZERO
from .contracts import Contracts, VersionContractNames


AdapterType = Literal[
    "aave_v2",
    "aave_v3",
    "alpha_homora_v1",
    "aura",
    "balancer_v2",
    "compound_v2",
    "compound_v3",
    "convex",
    "curve_exchange",
    "curve_liquidity",
    "curve_liquidity_aave",
    "curve_liquidity_eurs",
    "curve_liquidity_seth",
    "curve_liquidity_steth",
    "erc_4626",
    "idle",
    "kyber_network",
    "olympus_v2",
    "one_inch_v5",
    "paraswap_v4",
    "paraswap_v5",
    "pendle_v2",
    "pooltogether_v4",
    "swell_staking",
    "synthetix",
    "three_one_third",
    "tracked_assets",
    "uniswap_v2",
    "uniswap_v2_liquidity",
    "uniswap_v3",
    "unknown",
    "yearn_vault_v2",
    "zeroex_v2",
    "zeroex_v4",
    "zeroex_v4_pmm_kyc",
    "zero_lend_lrt_btc_aave_v3",
    "zero_lend_rwa_stablecoins_aave_v3",
]

KnownAdapterType = Literal[
    "aave_v2",
    "aave_v3",
    "alpha_homora_v1",
    "aura",
    "balancer_v2",
    "compound_v2",
    "compound_v3",
    "convex",
    "curve_exchange",
    "curve_liquidity",
    "curve_liquidity_aave",
    "curve_liquidity_eurs",
    "curve_liquidity_seth",
    "curve_liquidity_steth",
    "erc_4626",
    "idle",
    "kyber_network",
    "olympus_v2",
    "one_inch_v5",
    "paraswap_v4",
    "paraswap_v5",
    "pendle_v2",
    "pooltogether_v4",
    "swell_staking",
    "synthetix",
    "three_one_third",
    "tracked_assets",
    "uniswap_v2",
    "uniswap_v2_liquidity",
    "uniswap_v3",
    "yearn_vault_v2",
    "zeroex_v2",
    "zeroex_v4",
    "zeroex_v4_pmm_kyc",
]

UnknownAdapterType = Literal["unknown"]


class UnresolvedAdapterDefinition(TypedDict):
    contract_name: VersionContractNames | None
    name: str
    type: AdapterType


class AdapterDefinition(TypedDict):
    address: ChecksumAddress
    name: str
    type: AdapterType


def is_adapter_definition(adapter: dict[str, Any]) -> bool:
    return (
        isinstance(adapter, dict)
        and "address" in adapter
        and "name" in adapter
        and "type" in adapter
        and Web3.is_checksum_address(adapter["address"])
        and isinstance(adapter["name"], str)
        and isinstance(adapter["type"], str)
        and adapter["type"] in get_args(AdapterType)
    )


def _is_in_contracts(
    contract_name: Any,
    contracts: Contracts,
) -> bool:
    return contract_name in contracts


def get_adapters_for_release(
    contracts: Contracts,
) -> dict[ChecksumAddress, AdapterDefinition]:
    result = {}

    for adapter_def in ADAPTER_DEFINITIONS.values():
        contract_name = adapter_def.get("contract_name")
        if contract_name is not None and _is_in_contracts(contract_name, contracts):
            address = contracts[contract_name]
            if address != ADDRESS_ZERO:
                result[address] = {
                    "address": address,
                    "name": adapter_def["name"],
                    "type": adapter_def["type"],
                }

    return result


KNOWN_ADAPTER_DEFINITIONS: dict[KnownAdapterType, UnresolvedAdapterDefinition] = {
    "aave_v2": {
        "contract_name": "AaveV2Adapter",
        "name": "Aave V2 Supply",
        "type": "aave_v2",  # AdapterType
    },
    "aave_v3": {
        "contract_name": "AaveV3Adapter",
        "name": "Aave V3 Supply",
        "type": "aave_v3",  # AdapterType
    },
    "alpha_homora_v1": {
        "contract_name": None,
        "name": "Alpha Homora V1",
        "type": "alpha_homora_v1",  # AdapterType
    },
    "aura": {
        "contract_name": None,
        "name": "Aura Stake",
        "type": "aura",  # AdapterType
    },
    "balancer_v2": {
        "contract_name": "BalancerV2LiquidityAdapter",
        "name": "Balancer V2",
        "type": "balancer_v2",  # AdapterType
    },
    "compound_v2": {
        "contract_name": "CompoundAdapter",
        "name": "Compound Lend",
        "type": "compound_v2",  # AdapterType
    },
    "compound_v3": {
        "contract_name": "CompoundV3Adapter",
        "name": "Compound V3 Lend",
        "type": "compound_v3",  # AdapterType
    },
    "convex": {
        "contract_name": None,
        "name": "Convex Stake",
        "type": "convex",  # AdapterType
    },
    "curve_exchange": {
        "contract_name": "CurveExchangeAdapter",
        "name": "Curve Swap",
        "type": "curve_exchange",  # AdapterType
    },
    "curve_liquidity": {
        "contract_name": "CurveLiquidityAdapter",
        "name": "Curve Provide Liquidity",
        "type": "curve_liquidity",  # AdapterType
    },
    "curve_liquidity_aave": {
        "contract_name": None,
        "name": "Curve Aave Pool",
        "type": "curve_liquidity_aave",  # AdapterType
    },
    "curve_liquidity_eurs": {
        "contract_name": None,
        "name": "Curve Eurs Pool",
        "type": "curve_liquidity_eurs",  # AdapterType
    },
    "curve_liquidity_seth": {
        "contract_name": None,
        "name": "Curve Seth Pool",
        "type": "curve_liquidity_seth",  # AdapterType
    },
    "curve_liquidity_steth": {
        "contract_name": None,
        "name": "Curve Steth Pool",
        "type": "curve_liquidity_steth",  # AdapterType
    },
    "erc_4626": {
        "contract_name": "ERC4626Adapter",
        "name": "ERC4626",
        "type": "erc_4626",  # AdapterType
    },
    "idle": {
        "contract_name": None,
        "name": "Idle",
        "type": "idle",  # AdapterType
    },
    "kyber_network": {
        "contract_name": None,
        "name": "Kyber Network",
        "type": "kyber_network",  # AdapterType
    },
    "olympus_v2": {
        "contract_name": None,
        "name": "Olympus DAO",
        "type": "olympus_v2",  # AdapterType
    },
    "one_inch_v5": {
        "contract_name": "OneInchV5Adapter",
        "name": "1inch V5",
        "type": "one_inch_v5",  # AdapterType
    },
    "paraswap_v4": {
        "contract_name": None,
        "name": "ParaSwap V4",
        "type": "paraswap_v4",  # AdapterType
    },
    "paraswap_v5": {
        "contract_name": "ParaSwapV5Adapter",
        "name": "ParaSwap V5",
        "type": "paraswap_v5",  # AdapterType
    },
    "pendle_v2": {
        "contract_name": "PendleV2Adapter",
        "name": "Pendle V2",
        "type": "pendle_v2",  # AdapterType
    },
    "pooltogether_v4": {
        "contract_name": "PoolTogetherV4Adapter",
        "name": "PoolTogether",
        "type": "pooltogether_v4",  # AdapterType
    },
    "swell_staking": {
        "contract_name": "SwellStakingAdapter",
        "name": "Swell Staking",
        "type": "swell_staking",  # AdapterType
    },
    "synthetix": {
        "contract_name": "SynthetixAdapter",
        "name": "Synthetix",
        "type": "synthetix",  # AdapterType
    },
    "three_one_third": {
        "contract_name": "ThreeOneThirdAdapter",
        "name": "31Third",
        "type": "three_one_third",  # AdapterType
    },
    "tracked_assets": {
        "contract_name": None,
        "name": "Tracked Asset",
        "type": "tracked_assets",  # AdapterType
    },
    "uniswap_v2": {
        "contract_name": "UniswapV2ExchangeAdapter",
        "name": "Uniswap V2 Swap",
        "type": "uniswap_v2",  # AdapterType
    },
    "uniswap_v2_liquidity": {
        "contract_name": "UniswapV2LiquidityAdapter",
        "name": "Uniswap V2 Provide Liquidity",
        "type": "uniswap_v2_liquidity",  # AdapterType
    },
    "uniswap_v3": {
        "contract_name": "UniswapV3Adapter",
        "name": "Uniswap V3 Swap",
        "type": "uniswap_v3",  # AdapterType
    },
    "yearn_vault_v2": {
        "contract_name": "YearnVaultV2Adapter",
        "name": "Yearn",
        "type": "yearn_vault_v2",  # AdapterType
    },
    "zeroex_v2": {
        "contract_name": "ZeroExV2Adapter",
        "name": "0x V2",
        "type": "zeroex_v2",  # AdapterType
    },
    "zeroex_v4": {
        "contract_name": "ZeroExV4Adapter",
        "name": "0x V4",
        "type": "zeroex_v4",  # AdapterType
    },
    "zeroex_v4_pmm_kyc": {
        "contract_name": "ZeroExV4AdapterPmm2Kyc",
        "name": "0x V4",
        "type": "zeroex_v4_pmm_kyc",  # AdapterType
    },
    "zero_lend_lrt_btc_aave_v3": {
        "contract_name": "ZeroLendLRTBTCAaveV3Adapter",
        "name": "Zero Lend LRT BTC Supply",
        "type": "zero_lend_lrt_btc_aave_v3",  # AdapterType
    },
    "zero_lend_rwa_stablecoins_aave_v3": {
        "contract_name": "ZeroLendRWAStablecoinsAaveV3Adapter",
        "name": "Zero Lend RWA Stablecoins Supply",
        "type": "zero_lend_rwa_stablecoins_aave_v3",  # AdapterType
    },
}

UNKNOWN_ADAPTER_DEFINITIONS: dict[UnknownAdapterType, UnresolvedAdapterDefinition] = {
    "unknown": {
        "contract_name": None,
        "name": "Unknown",
        "type": "unknown",  # AdapterType
    },
}

ADAPTER_DEFINITIONS: dict[AdapterType, UnresolvedAdapterDefinition] = {
    **KNOWN_ADAPTER_DEFINITIONS,
    **UNKNOWN_ADAPTER_DEFINITIONS,
}



from typing import TypedDict
from web3.types import ChecksumAddress
from .utils.clients import PublicClient
from .configuration_lib import policy
from ..abis import abis


async def get_enabled_policies(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
    policy_manager: ChecksumAddress,
) -> list[ChecksumAddress]:
    contract = client.contract(policy_manager, abis.IPolicyManager)
    function = contract.functions.getEnabledPoliciesForFund(comptroller_proxy)
    return await function.call()



import json
from pathlib import Path
from typing import Any

_abis: dict[str, Any] = {}

base_path = Path(__file__).parent
abis_path = base_path / "abis"

def __getattr__(name: str) -> Any:
    if name in _abis:
        return _abis[name]

    json_file = abis_path / f"{name}.abi.json"
    if json_file.is_file():
        try:
            with open(json_file, "r") as f:
                _abis[name] = json.load(f)
            return _abis[name]
        except json.JSONDecodeError:
            raise ImportError(f"Invalid JSON in {json_file}")
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

__all__ = [f.stem.split(".abi")[0] for f in abis_path.glob("*.abi.json")]


import asyncio
from decimal import Decimal
from eth_abi import encode, decode
from web3.types import ChecksumAddress, HexStr, TxParams
from web3 import Web3
from typing import Tuple, TypedDict
from ..._internal import integration_manager as integration_manager_lib
from ..._internal import external_position_manager as external_position_manager_lib
from ...utils.clients import PublicClient
from ...utils.conversion import from_wei
from ...utils.encoding import encoding_to_types
from ... import asset as asset_lib


# --------------------------------------------------------------------------------------------
# LEND
# --------------------------------------------------------------------------------------------


async def lend(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, ChecksumAddress | int]
            {
                "a_token": ChecksumAddress,
                "deposit_amount": int,
            }
    """
    _lend = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["lend"], lend_encode
    )
    return await _lend(*args)


LEND_ENCODING = [
    {
        "type": "address",
        "name": "aToken",
    },
    {
        "type": "uint256",
        "name": "depositAmount",
    },
]


class LendArgs(TypedDict):
    a_token: ChecksumAddress
    deposit_amount: int


def lend_encode(a_token: ChecksumAddress, deposit_amount: int) -> HexStr:
    types = encoding_to_types(LEND_ENCODING)
    values = [a_token, deposit_amount]
    return Web3.to_hex(encode(types, values))


def lend_decode(encoded: HexStr) -> dict[str, ChecksumAddress | int]:
    """
    Returns:
        {
            "a_token": ChecksumAddress,
            "deposit_amount": int,
        }
    """
    types = encoding_to_types(LEND_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "a_token": Web3.to_checksum_address(decoded[0]),
        "deposit_amount": decoded[1],
    }


# --------------------------------------------------------------------------------------------
# REDEEM
# --------------------------------------------------------------------------------------------


async def redeem(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, ChecksumAddress | int]
            {
                "a_token": ChecksumAddress,
                "redeem_amount": int,
            }
    """
    _redeem = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["redeem"], redeem_encode
    )
    return await _redeem(*args)


REDEEM_ENCODING = [
    {
        "type": "address",
        "name": "aToken",
    },
    {
        "type": "uint256",
        "name": "redeemAmount",
    },
]


class RedeemArgs(TypedDict):
    a_token: ChecksumAddress
    redeem_amount: int


def redeem_encode(a_token: ChecksumAddress, redeem_amount: int) -> HexStr:
    types = encoding_to_types(REDEEM_ENCODING)
    values = [a_token, redeem_amount]
    return Web3.to_hex(encode(types, values))


def redeem_decode(encoded: HexStr) -> dict[str, ChecksumAddress | int]:
    """
    Returns:
        {
            "a_token": ChecksumAddress,
            "redeem_amount": int,
        }
    """
    types = encoding_to_types(REDEEM_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "a_token": Web3.to_checksum_address(decoded[0]),
        "redeem_amount": decoded[1],
    }


# --------------------------------------------------------------------------------------------
# EXTERNAL POSITION
# --------------------------------------------------------------------------------------------


class Action(TypedDict):
    add_collateral: int
    remove_collateral: int
    borrow: int
    repay_borrow: int
    set_e_mode: int
    set_use_reserve_as_collateral: int
    claim_rewards: int
    sweep: int


ACTION = {
    "add_collateral": 0,
    "remove_collateral": 1,
    "borrow": 2,
    "repay_borrow": 3,
    "set_e_mode": 4,
    "set_use_reserve_as_collateral": 5,
    "claim_rewards": 6,
    "sweep": 7,
}

create = external_position_manager_lib.create_only


# --------------------------------------------------------------------------------------------
# ADD COLLATERAL
# --------------------------------------------------------------------------------------------


async def add_collateral(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress] | list[int] | bool]
            {
                "a_tokens": list[ChecksumAddress],
                "amounts": list[int],
                "from_underlying": bool,
            }
    """
    _add_collateral = external_position_manager_lib.make_use(
        ACTION["add_collateral"], add_collateral_encode
    )
    return await _add_collateral(*args)


async def create_and_add_collateral(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        type_id: int
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        initialization_data: HexStr
        call_args: dict[str, list[ChecksumAddress] | list[int] | bool]
            {
                "a_tokens": list[ChecksumAddress],
                "amounts": list[int],
                "from_underlying": bool,
            }
    """
    _create_and_add_collateral = external_position_manager_lib.make_create_and_use(
        ACTION["add_collateral"], add_collateral_encode
    )
    return await _create_and_add_collateral(*args)


ADD_COLLATERAL_ENCODING = [
    {
        "type": "address[]",
        "name": "aTokens",
    },
    {
        "type": "uint256[]",
        "name": "amounts",
    },
    {
        "type": "bool",
        "name": "fromUnderlying",
    },
]


class AddCollateralArgs(TypedDict):
    a_tokens: list[ChecksumAddress]
    amounts: list[int]
    from_underlying: bool


def add_collateral_encode(
    a_tokens: list[ChecksumAddress], amounts: list[int], from_underlying: bool
) -> HexStr:
    types = encoding_to_types(ADD_COLLATERAL_ENCODING)
    values = [a_tokens, amounts, from_underlying]
    return Web3.to_hex(encode(types, values))


def add_collateral_decode(
    encoded: HexStr,
) -> dict[str, list[ChecksumAddress] | list[int] | bool]:
    """
    Returns:
        {
            "a_tokens": list[ChecksumAddress],
            "amounts": list[int],
            "from_underlying": bool,
        }
    """
    types = encoding_to_types(ADD_COLLATERAL_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "a_tokens": [Web3.to_checksum_address(a_token) for a_token in decoded[0]],
        "amounts": decoded[1],
        "from_underlying": decoded[2],
    }


# --------------------------------------------------------------------------------------------
# REMOVE COLLATERAL
# --------------------------------------------------------------------------------------------


async def remove_collateral(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress] | list[int] | bool]
            {
                "a_tokens": list[ChecksumAddress],
                "amounts": list[int],
                "to_underlying": bool,
            }
    """
    _remove_collateral = external_position_manager_lib.make_use(
        ACTION["remove_collateral"], remove_collateral_encode
    )
    return await _remove_collateral(*args)


REMOVE_COLLATERAL_ENCODING = [
    {
        "type": "address[]",
        "name": "aTokens",
    },
    {
        "type": "uint256[]",
        "name": "amounts",
    },
    {
        "type": "bool",
        "name": "toUnderlying",
    },
]


class RemoveCollateralArgs(TypedDict):
    a_tokens: list[ChecksumAddress]
    amounts: list[int]
    to_underlying: bool


def remove_collateral_encode(
    a_tokens: list[ChecksumAddress], amounts: list[int], to_underlying: bool
) -> HexStr:
    types = encoding_to_types(REMOVE_COLLATERAL_ENCODING)
    values = [a_tokens, amounts, to_underlying]
    return Web3.to_hex(encode(types, values))


def remove_collateral_decode(
    encoded: HexStr,
) -> dict[str, list[ChecksumAddress] | list[int] | bool]:
    """
    Returns:
        {
            "a_tokens": list[ChecksumAddress],
            "amounts": list[int],
            "to_underlying": bool,
        }
    """
    types = encoding_to_types(REMOVE_COLLATERAL_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "a_tokens": [Web3.to_checksum_address(a_token) for a_token in decoded[0]],
        "amounts": decoded[1],
        "to_underlying": decoded[2],
    }


# --------------------------------------------------------------------------------------------
# BORROW
# --------------------------------------------------------------------------------------------


async def borrow(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress] | list[int]]
            {
                "underlying_tokens": list[ChecksumAddress],
                "amounts": list[int],
            }
    """
    _borrow = external_position_manager_lib.make_use(ACTION["borrow"], borrow_encode)
    return await _borrow(*args)


async def create_and_borrow(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        type_id: int
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        initialization_data: HexStr
        call_args: dict[str, list[ChecksumAddress] | list[int]]
            {
                "underlying_tokens": list[ChecksumAddress],
                "amounts": list[int],
            }
    """
    _create_and_borrow = external_position_manager_lib.make_create_and_use(
        ACTION["borrow"], borrow_encode
    )
    return await _create_and_borrow(*args)


BORROW_ENCODING = [
    {
        "type": "address[]",
        "name": "underlyingTokens",
    },
    {
        "type": "uint256[]",
        "name": "amounts",
    },
]


class BorrowArgs(TypedDict):
    underlying_tokens: list[ChecksumAddress]
    amounts: list[int]


def borrow_encode(
    underlying_tokens: list[ChecksumAddress], amounts: list[int]
) -> HexStr:
    types = encoding_to_types(BORROW_ENCODING)
    values = [underlying_tokens, amounts]
    return Web3.to_hex(encode(types, values))


def borrow_decode(encoded: HexStr) -> dict[str, list[ChecksumAddress] | list[int]]:
    """
    Returns:
        {
            "underlying_tokens": list[ChecksumAddress],
            "amounts": list[int],
        }
    """
    types = encoding_to_types(BORROW_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "underlying_tokens": [
            Web3.to_checksum_address(underlying_token)
            for underlying_token in decoded[0]
        ],
        "amounts": decoded[1],
    }


# --------------------------------------------------------------------------------------------
# REPAY BORROW
# --------------------------------------------------------------------------------------------


async def repay_borrow(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress] | list[int]]
            {
                "underlying_tokens": list[ChecksumAddress],
                "amounts": list[int],
            }
    """
    _repay_borrow = external_position_manager_lib.make_use(
        ACTION["repay_borrow"], repay_borrow_encode
    )
    return await _repay_borrow(*args)


REPAY_BORROW_ENCODING = [
    {
        "type": "address[]",
        "name": "underlyingTokens",
    },
    {
        "type": "uint256[]",
        "name": "amounts",
    },
]


class RepayBorrowArgs(TypedDict):
    underlying_tokens: list[ChecksumAddress]
    amounts: list[int]


def repay_borrow_encode(
    underlying_tokens: list[ChecksumAddress], amounts: list[int]
) -> HexStr:
    types = encoding_to_types(REPAY_BORROW_ENCODING)
    values = [underlying_tokens, amounts]
    return Web3.to_hex(encode(types, values))


def repay_borrow_decode(
    encoded: HexStr,
) -> dict[str, list[ChecksumAddress] | list[int]]:
    """
    Returns:
        {
            "underlying_tokens": list[ChecksumAddress],
            "amounts": list[int],
        }
    """
    types = encoding_to_types(REPAY_BORROW_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "underlying_tokens": [
            Web3.to_checksum_address(underlying_token)
            for underlying_token in decoded[0]
        ],
        "amounts": decoded[1],
    }


# --------------------------------------------------------------------------------------------
# SET E-MODE
# --------------------------------------------------------------------------------------------


async def set_e_mode(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, int]
            {
                "category_id": int,
            }
    """
    _set_e_mode = external_position_manager_lib.make_use(
        ACTION["set_e_mode"], set_e_mode_encode
    )
    return await _set_e_mode(*args)


SET_EMODE_ENCODING = [
    {
        "type": "uint8",
        "name": "categoryId",
    },
]


class SetEModeArgs(TypedDict):
    category_id: int


def set_e_mode_encode(category_id: int) -> HexStr:
    types = encoding_to_types(SET_EMODE_ENCODING)
    values = [category_id]
    return Web3.to_hex(encode(types, values))


def set_e_mode_decode(encoded: HexStr) -> dict[str, int]:
    """
    Returns:
        {
            "category_id": int,
        }
    """
    types = encoding_to_types(SET_EMODE_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "category_id": decoded[0],
    }


# --------------------------------------------------------------------------------------------
# SET USE RESERVE AS COLLATERAL
# --------------------------------------------------------------------------------------------


async def set_use_reserve_as_collateral(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, ChecksumAddress | bool]
            {
                "underlying": ChecksumAddress,
                "use_as_collateral": bool,
            }
    """
    _set_use_reserve_as_collateral = external_position_manager_lib.make_use(
        ACTION["set_use_reserve_as_collateral"], set_use_reserve_as_collateral_encode
    )
    return await _set_use_reserve_as_collateral(*args)


SET_USE_RESERVE_AS_COLLATERAL_ENCODING = [
    {
        "type": "address",
        "name": "underlying",
    },
    {
        "type": "bool",
        "name": "useAsCollateral",
    },
]


class SetUseReserveAsCollateralArgs(TypedDict):
    underlying: ChecksumAddress
    use_as_collateral: bool


def set_use_reserve_as_collateral_encode(
    underlying: ChecksumAddress, use_as_collateral: bool
) -> HexStr:
    types = encoding_to_types(SET_USE_RESERVE_AS_COLLATERAL_ENCODING)
    values = [underlying, use_as_collateral]
    return Web3.to_hex(encode(types, values))


def set_use_reserve_as_collateral_decode(
    encoded: HexStr,
) -> dict[str, ChecksumAddress | bool]:
    """
    Returns:
        {
            "underlying": ChecksumAddress,
            "use_as_collateral": bool,
        }
    """
    types = encoding_to_types(SET_USE_RESERVE_AS_COLLATERAL_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "underlying": decoded[0],
        "use_as_collateral": decoded[1],
    }


# --------------------------------------------------------------------------------------------
# CLAIM REWARDS
# --------------------------------------------------------------------------------------------


async def claim_rewards(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress] | int | ChecksumAddress]
            {
                "assets": list[ChecksumAddress],
                "amount": int,
                "reward_token": ChecksumAddress,
            }
    """
    _claim_rewards = external_position_manager_lib.make_use(
        ACTION["claim_rewards"], claim_rewards_encode
    )
    return await _claim_rewards(*args)


CLAIM_REWARDS_ENCODING = [
    {
        "type": "address[]",
        "name": "assets",
    },
    {
        "type": "uint256",
        "name": "amount",
    },
    {
        "type": "address",
        "name": "rewardToken",
    },
]


class ClaimRewardsArgs(TypedDict):
    assets: list[ChecksumAddress]
    amount: int
    reward_token: ChecksumAddress


def claim_rewards_encode(
    assets: list[ChecksumAddress], amount: int, reward_token: ChecksumAddress
) -> HexStr:
    types = encoding_to_types(CLAIM_REWARDS_ENCODING)
    values = [assets, amount, reward_token]
    return Web3.to_hex(encode(types, values))


def claim_rewards_decode(
    encoded: HexStr,
) -> dict[str, list[ChecksumAddress] | int | ChecksumAddress]:
    """
    Returns:
        {
            "assets": list[ChecksumAddress],
            "amount": int,
            "reward_token": ChecksumAddress,
        }
    """
    types = encoding_to_types(CLAIM_REWARDS_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "assets": [Web3.to_checksum_address(asset) for asset in decoded[0]],
        "amount": decoded[1],
        "reward_token": Web3.to_checksum_address(decoded[2]),
    }


# --------------------------------------------------------------------------------------------
# SWEEP
# --------------------------------------------------------------------------------------------


async def sweep(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        external_position_manager: ChecksumAddress
        external_position_proxy: ChecksumAddress
        call_args: dict[str, list[ChecksumAddress]]
            {
                "assets": list[ChecksumAddress],
            }
    """
    _sweep = external_position_manager_lib.make_use(ACTION["sweep"], sweep_encode)
    return await _sweep(*args)


SWEEP_ENCODING = [
    {
        "type": "address[]",
        "name": "assets",
    },
]


class SweepArgs(TypedDict):
    assets: list[ChecksumAddress]


def sweep_encode(assets: list[ChecksumAddress]) -> HexStr:
    types = encoding_to_types(SWEEP_ENCODING)
    values = [assets]
    return Web3.to_hex(encode(types, values))


def sweep_decode(encoded: HexStr) -> dict[str, list[ChecksumAddress]]:
    """
    Returns:
        {
            "assets": list[ChecksumAddress],
        }
    """
    types = encoding_to_types(SWEEP_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "assets": [Web3.to_checksum_address(asset) for asset in decoded[0]],
    }


# --------------------------------------------------------------------------------------------
# THIRD PARTY READ FUNCTIONS
# --------------------------------------------------------------------------------------------


POOL_ADDRESS_PROVIDER_ABI = [
    {
        "inputs": [],
        "name": "getPool",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
]


async def get_pool(
    client: PublicClient,
    pool_address_provider: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(pool_address_provider, POOL_ADDRESS_PROVIDER_ABI)
    function = contract.functions.getPool()
    return await function.call()


POOL_ABI = [
    {
        "inputs": [{"internalType": "uint8", "name": "id", "type": "uint8"}],
        "name": "getEModeCategoryData",
        "outputs": [
            {
                "components": [
                    {"internalType": "uint16", "name": "ltv", "type": "uint16"},
                    {
                        "internalType": "uint16",
                        "name": "liquidationThreshold",
                        "type": "uint16",
                    },
                    {
                        "internalType": "uint16",
                        "name": "liquidationBonus",
                        "type": "uint16",
                    },
                    {
                        "internalType": "address",
                        "name": "priceSource",
                        "type": "address",
                    },
                    {"internalType": "string", "name": "label", "type": "string"},
                ],
                "internalType": "struct DataTypes.EModeCategory",
                "name": "",
                "type": "tuple",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "user", "type": "address"}],
        "name": "getUserAccountData",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "totalCollateralBase",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "totalDebtBase", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "availableBorrowsBase",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "currentLiquidationThreshold",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "ltv", "type": "uint256"},
            {"internalType": "uint256", "name": "healthFactor", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
]


async def get_e_mode_category_data(
    client: PublicClient,
    pool: ChecksumAddress,
    category_id: int,
) -> Tuple[int, int, int, ChecksumAddress, str]:
    """
    Returns:
        (
            ltv: int,
            liquidation_threshold: int,
            liquidation_bonus: int,
            price_source: ChecksumAddress,
            label: str,
        )
    """
    contract = client.contract(pool, POOL_ABI)
    function = contract.functions.getEModeCategoryData(category_id)
    return await function.call()


async def get_user_account_data(
    client: PublicClient,
    pool: ChecksumAddress,
    user: ChecksumAddress,
) -> Tuple[int, int, int, int, int, int]:
    """
    Returns:
        (
            total_collateral_base: int,
            total_debt_base: int,
            available_borrows_base: int,
            current_liquidation_threshold: int,
            ltv: int,
            health_factor: int,
        )
    """
    contract = client.contract(pool, POOL_ABI)
    function = contract.functions.getUserAccountData(user)
    return await function.call()


REWARDS_CONTROLLER_ABI = [
    {
        "inputs": [
            {"internalType": "address[]", "name": "assets", "type": "address[]"},
            {"internalType": "address", "name": "user", "type": "address"},
        ],
        "name": "getAllUserRewards",
        "outputs": [
            {"internalType": "address[]", "name": "rewardsList", "type": "address[]"},
            {
                "internalType": "uint256[]",
                "name": "unclaimedAmounts",
                "type": "uint256[]",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "asset", "type": "address"}],
        "name": "getRewardsByAsset",
        "outputs": [{"internalType": "address[]", "name": "", "type": "address[]"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "asset", "type": "address"},
            {"internalType": "address", "name": "reward", "type": "address"},
        ],
        "name": "getRewardsData",
        "outputs": [
            {"internalType": "uint256", "name": "index", "type": "uint256"},
            {"internalType": "uint256", "name": "emissionPerSecond", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "lastUpdateTimestamp",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "distributionEnd", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
]


async def get_all_user_rewards(
    client: PublicClient,
    rewards_controller: ChecksumAddress,
    assets: list[ChecksumAddress],
    user: ChecksumAddress,
) -> Tuple[list[ChecksumAddress], list[int]]:
    """
    Returns:
        (
            rewards_list: list[ChecksumAddress],
            unclaimed_amounts: list[int],
        )
    """
    contract = client.contract(rewards_controller, REWARDS_CONTROLLER_ABI)
    function = contract.functions.getAllUserRewards(assets, user)
    return await function.call()


async def get_rewards_by_asset(
    client: PublicClient,
    rewards_controller: ChecksumAddress,
    asset: ChecksumAddress,
) -> list[ChecksumAddress]:
    """
    Returns:
        rewards_list: list[ChecksumAddress]
    """
    contract = client.contract(rewards_controller, REWARDS_CONTROLLER_ABI)
    function = contract.functions.getRewardsByAsset(asset)
    return await function.call()


async def get_rewards_data(
    client: PublicClient,
    rewards_controller: ChecksumAddress,
    asset: ChecksumAddress,
    reward: ChecksumAddress,
) -> Tuple[int, int, int, int]:
    """
    Returns:
        (
            index: int,
            emission_per_second: int,
            last_update_timestamp: int,
            distribution_end: int,
        )
    """
    contract = client.contract(rewards_controller, REWARDS_CONTROLLER_ABI)
    function = contract.functions.getRewardsData(asset, reward)
    return await function.call()


PROTOCOL_DATA_PROVIDER_ABI = [
    {
        "inputs": [{"internalType": "address", "name": "asset", "type": "address"}],
        "name": "getReserveCaps",
        "outputs": [
            {"internalType": "uint256", "name": "supplyCap", "type": "uint256"},
            {"internalType": "uint256", "name": "borrowCap", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "asset", "type": "address"}],
        "name": "getReserveData",
        "outputs": [
            {"internalType": "uint256", "name": "unbacked", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "accruedToTreasuryScaled",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "totalAToken", "type": "uint256"},
            {"internalType": "uint256", "name": "totalStableDebt", "type": "uint256"},
            {"internalType": "uint256", "name": "totalVariableDebt", "type": "uint256"},
            {"internalType": "uint256", "name": "liquidityRate", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "variableBorrowRate",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "stableBorrowRate", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "averageStableBorrowRate",
                "type": "uint256",
            },
            {"internalType": "uint256", "name": "liquidityIndex", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "variableBorrowIndex",
                "type": "uint256",
            },
            {"internalType": "uint40", "name": "lastUpdateTimestamp", "type": "uint40"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
]


async def get_reserve_caps(
    client: PublicClient,
    protocol_data_provider: ChecksumAddress,
    asset: ChecksumAddress,
) -> Tuple[Decimal, Decimal]:
    """
    Returns:
        (
            supply_cap: Decimal,
            borrow_cap: Decimal,
        )
    """
    contract = client.contract(protocol_data_provider, PROTOCOL_DATA_PROVIDER_ABI)
    function = contract.functions.getReserveCaps(asset)
    reserve_caps, decimals = await asyncio.gather(
        function.call(),
        asset_lib.get_decimals(client, asset),
    )
    return (
        from_wei(reserve_caps[0], decimals),
        from_wei(reserve_caps[1], decimals),
    )


async def get_reserve_data(
    client: PublicClient,
    protocol_data_provider: ChecksumAddress,
    asset: ChecksumAddress,
) -> Tuple[int, int, int, int, int, int, int, int, int, int, int]:
    """
    Returns:
        (
            unbacked: int,
            accrued_to_treasury_scaled: int,
            total_a_token: int,
            total_stable_debt: int,
            total_variable_debt: int,
            liquidity_rate: int,
            variable_borrow_rate: int,
            stable_borrow_rate: int,
            average_stable_borrow_rate: int,
            liquidity_index: int,
            variable_borrow_index: int,
            last_update_timestamp: int,
        )
    """
    contract = client.contract(protocol_data_provider, PROTOCOL_DATA_PROVIDER_ABI)
    function = contract.functions.getReserveData(asset)
    return await function.call()


async def get_available_supply_amount(
    client: PublicClient,
    protocol_data_provider: ChecksumAddress,
    asset: ChecksumAddress,
    decimals: int,
) -> Decimal:
    reserve_caps, reserve_data = await asyncio.gather(
        get_reserve_caps(client, protocol_data_provider, asset),
        get_reserve_data(client, protocol_data_provider, asset),
    )
    return reserve_caps[1] - from_wei(reserve_data[2], decimals)


async def get_available_variable_debt_amount(
    client: PublicClient,
    protocol_data_provider: ChecksumAddress,
    asset: ChecksumAddress,
    decimals: int,
) -> Decimal:
    reserve_caps, reserve_data = await asyncio.gather(
        get_reserve_caps(client, protocol_data_provider, asset),
        get_reserve_data(client, protocol_data_provider, asset),
    )
    return reserve_caps[0] - from_wei(reserve_data[4], decimals)


UI_INCENTIVE_DATA_PROVIDER_ABI = [
    {
        "inputs": [{"internalType": "address", "name": "provider", "type": "address"}],
        "name": "getReservesIncentivesData",
        "outputs": [
            {
                "components": [
                    {
                        "internalType": "address",
                        "name": "underlyingAsset",
                        "type": "address",
                    },
                    {
                        "components": [
                            {
                                "internalType": "address",
                                "name": "tokenAddress",
                                "type": "address",
                            },
                            {
                                "internalType": "address",
                                "name": "incentiveControllerAddress",
                                "type": "address",
                            },
                            {
                                "components": [
                                    {
                                        "internalType": "string",
                                        "name": "rewardTokenSymbol",
                                        "type": "string",
                                    },
                                    {
                                        "internalType": "address",
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {
                                        "internalType": "address",
                                        "name": "rewardOracleAddress",
                                        "type": "address",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "emissionPerSecond",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "incentivesLastUpdateTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "tokenIncentivesIndex",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "emissionEndTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "int256",
                                        "name": "rewardPriceFeed",
                                        "type": "int256",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "rewardTokenDecimals",
                                        "type": "uint8",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "precision",
                                        "type": "uint8",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "priceFeedDecimals",
                                        "type": "uint8",
                                    },
                                ],
                                "internalType": "struct IUiIncentiveDataProviderV3.RewardInfo[]",
                                "name": "rewardsTokenInformation",
                                "type": "tuple[]",
                            },
                        ],
                        "internalType": "struct IUiIncentiveDataProviderV3.IncentiveData",
                        "name": "aIncentiveData",
                        "type": "tuple",
                    },
                    {
                        "components": [
                            {
                                "internalType": "address",
                                "name": "tokenAddress",
                                "type": "address",
                            },
                            {
                                "internalType": "address",
                                "name": "incentiveControllerAddress",
                                "type": "address",
                            },
                            {
                                "components": [
                                    {
                                        "internalType": "string",
                                        "name": "rewardTokenSymbol",
                                        "type": "string",
                                    },
                                    {
                                        "internalType": "address",
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {
                                        "internalType": "address",
                                        "name": "rewardOracleAddress",
                                        "type": "address",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "emissionPerSecond",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "incentivesLastUpdateTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "tokenIncentivesIndex",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "uint256",
                                        "name": "emissionEndTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "internalType": "int256",
                                        "name": "rewardPriceFeed",
                                        "type": "int256",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "rewardTokenDecimals",
                                        "type": "uint8",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "precision",
                                        "type": "uint8",
                                    },
                                    {
                                        "internalType": "uint8",
                                        "name": "priceFeedDecimals",
                                        "type": "uint8",
                                    },
                                ],
                                "internalType": "struct IUiIncentiveDataProviderV3.RewardInfo[]",
                                "name": "rewardsTokenInformation",
                                "type": "tuple[]",
                            },
                        ],
                        "internalType": "struct IUiIncentiveDataProviderV3.IncentiveData",
                        "name": "vIncentiveData",
                        "type": "tuple",
                    },
                ],
                "internalType": "struct IUiIncentiveDataProviderV3.AggregatedReserveIncentiveData[]",
                "name": "",
                "type": "tuple[]",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
]


async def get_reserves_incentives_data(
    client: PublicClient,
    ui_incentive_data_provider: ChecksumAddress,
    pool_address_provider: ChecksumAddress,
):
    # TODO: unknown return type
    contract = client.contract(
        ui_incentive_data_provider, UI_INCENTIVE_DATA_PROVIDER_ABI
    )
    function = contract.functions.getReservesIncentivesData(pool_address_provider)
    return await function.call()



from typing import TypedDict
from web3 import Web3
from web3.types import ChecksumAddress, TxParams, HexStr
from web3.constants import ADDRESS_ZERO
from .utils.clients import PublicClient, WalletClient
from ..abis import abis


RELAY_HUB_ABI = [
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "maxAcceptanceBudget",
                "type": "uint256",
            },
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "address",
                                "name": "from",
                                "type": "address",
                            },
                            {
                                "internalType": "address",
                                "name": "to",
                                "type": "address",
                            },
                            {
                                "internalType": "uint256",
                                "name": "value",
                                "type": "uint256",
                            },
                            {
                                "internalType": "uint256",
                                "name": "gas",
                                "type": "uint256",
                            },
                            {
                                "internalType": "uint256",
                                "name": "nonce",
                                "type": "uint256",
                            },
                            {"internalType": "bytes", "name": "data", "type": "bytes"},
                            {
                                "internalType": "uint256",
                                "name": "validUntil",
                                "type": "uint256",
                            },
                        ],
                        "internalType": "struct IForwarder.ForwardRequest",
                        "name": "request",
                        "type": "tuple",
                    },
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "gasPrice",
                                "type": "uint256",
                            },
                            {
                                "internalType": "uint256",
                                "name": "pctRelayFee",
                                "type": "uint256",
                            },
                            {
                                "internalType": "uint256",
                                "name": "baseRelayFee",
                                "type": "uint256",
                            },
                            {
                                "internalType": "address",
                                "name": "relayWorker",
                                "type": "address",
                            },
                            {
                                "internalType": "address",
                                "name": "paymaster",
                                "type": "address",
                            },
                            {
                                "internalType": "address",
                                "name": "forwarder",
                                "type": "address",
                            },
                            {
                                "internalType": "bytes",
                                "name": "paymasterData",
                                "type": "bytes",
                            },
                            {
                                "internalType": "uint256",
                                "name": "clientId",
                                "type": "uint256",
                            },
                        ],
                        "internalType": "struct GsnTypes.RelayData",
                        "name": "relayData",
                        "type": "tuple",
                    },
                ],
                "internalType": "struct GsnTypes.RelayRequest",
                "name": "relayRequest",
                "type": "tuple",
            },
            {"internalType": "bytes", "name": "signature", "type": "bytes"},
            {"internalType": "bytes", "name": "approvalData", "type": "bytes"},
            {"internalType": "uint256", "name": "externalGasLimit", "type": "uint256"},
        ],
        "name": "relayCall",
        "outputs": [
            {"internalType": "bool", "name": "paymasterAccepted", "type": "bool"},
            {"internalType": "bytes", "name": "returnValue", "type": "bytes"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]

RELAY_REQUEST_TYPES = {
    "RelayData": [
        {"name": "gasPrice", "type": "uint256"},
        {"name": "pctRelayFee", "type": "uint256"},
        {"name": "baseRelayFee", "type": "uint256"},
        {"name": "relayWorker", "type": "address"},
        {"name": "paymaster", "type": "address"},
        {"name": "forwarder", "type": "address"},
        {"name": "paymasterData", "type": "bytes"},
        {"name": "clientId", "type": "uint256"},
    ],
    "RelayRequest": [
        {"name": "from", "type": "address"},
        {"name": "to", "type": "address"},
        {"name": "value", "type": "uint256"},
        {"name": "gas", "type": "uint256"},
        {"name": "nonce", "type": "uint256"},
        {"name": "data", "type": "bytes"},
        {"name": "validUntil", "type": "uint256"},
        {"name": "relayData", "type": "RelayData"},
    ],
}


# --------------------------------------------------------------------------------------------
# TRANSACTIONS
# --------------------------------------------------------------------------------------------


async def deploy_gas_relay_paymaster(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.deployGasRelayPaymaster()
    return await client.populated_transaction(function)


async def deposit_to_gas_relay_paymaster(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.depositToGasRelayPaymaster()
    return await client.populated_transaction(function)


async def shutdown_gas_relay_paymaster(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.shutdownGasRelayPaymaster()
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# READ FUNCTIONS
# --------------------------------------------------------------------------------------------


async def is_relayer_enabled(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> bool:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getGasRelayPaymaster()
    gas_relay_paymaster_address = await function.call()
    return gas_relay_paymaster_address != ADDRESS_ZERO


async def get_gas_relay_paymaster(
    client: PublicClient,
    comptroller_proxy: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.getGasRelayPaymaster()
    return await function.call()


async def get_relayer_balance(
    client: PublicClient,
    gas_relay_paymaster: ChecksumAddress,
) -> int:
    contract = client.contract(gas_relay_paymaster, abis.IGasRelayPaymasterLib)
    function = contract.functions.getRelayHubDeposit()
    return await function.call()


async def get_trusted_forwarder(
    client: PublicClient,
    gas_relay_paymaster: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(gas_relay_paymaster, abis.IGasRelayPaymasterLib)
    function = contract.functions.trustedForwarder()
    return await function.call()


async def get_nonce(
    client: PublicClient,
    trusted_forwarder: ChecksumAddress,
    sender: ChecksumAddress,
) -> int:
    contract = client.contract(trusted_forwarder, abis.IForwarder)
    function = contract.functions.getNonce(sender)
    return await function.call()


class Request(TypedDict):
    from_: ChecksumAddress
    to: ChecksumAddress
    value: int
    gas: int
    nonce: int
    data: HexStr
    validUntil: int


class RelayData(TypedDict):
    gasPrice: int
    pctRelayFee: int
    baseRelayFee: int
    relayWorker: ChecksumAddress
    paymaster: ChecksumAddress
    forwarder: ChecksumAddress
    paymasterData: HexStr
    clientId: int


class RelayRequest(TypedDict):
    request: Request
    relayData: RelayData


def encode_relay_call_data(
    max_acceptance_budget: int,
    relay_request: RelayRequest,
    signature: HexStr,
    approval_data: HexStr,
    gas_limit: int,
) -> HexStr:
    """
    Args:
        relay_request:
            {
                "request": {
                    "from": ChecksumAddress,
                    "to": ChecksumAddress,
                    "value": int,
                    "gas": int,
                    "nonce": int,
                    "data": HexStr,
                    "validUntil": int,
                },
                "relayData": {
                    "gasPrice": int,
                    "pctRelayFee": int,
                    "baseRelayFee": int,
                    "relayWorker": ChecksumAddress,
                    "paymaster": ChecksumAddress,
                    "forwarder": ChecksumAddress,
                    "paymasterData": HexStr,
                }
            }
    """
    contract = Web3().eth.contract(abi=RELAY_HUB_ABI)
    return contract.encode_abi(
        "relayCall",
        [
            max_acceptance_budget,
            relay_request,
            signature,
            approval_data,
            gas_limit,
        ],
    )
